# Copyright 2019 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""A wrapper around the Keras Resnet V1 models for object detection."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import tensorflow.compat.v1 as tf

from tensorflow.python.keras.applications import resnet
from object_detection.core import freezable_batch_norm
from object_detection.models.keras_models import model_utils


def _fixed_padding(inputs, kernel_size, rate=1):  # pylint: disable=invalid-name
  """Pads the input along the spatial dimensions independently of input size.

  Pads the input such that if it was used in a convolution with 'VALID' padding,
  the output would have the same dimensions as if the unpadded input was used
  in a convolution with 'SAME' padding.

  Args:
    inputs: A tensor of size [batch, height_in, width_in, channels].
    kernel_size: The kernel to be used in the conv2d or max_pool2d operation.
    rate: An integer, rate for atrous convolution.

  Returns:
    output: A tensor of size [batch, height_out, width_out, channels] with the
      input, either intact (if kernel_size == 1) or padded (if kernel_size > 1).
  """
  kernel_size_effective = kernel_size + (kernel_size - 1) * (rate - 1)
  pad_total = kernel_size_effective - 1
  pad_beg = pad_total // 2
  pad_end = pad_total - pad_beg
  padded_inputs = tf.pad(
      inputs, [[0, 0], [pad_beg, pad_end], [pad_beg, pad_end], [0, 0]])
  return padded_inputs


class _LayersOverride(object):
  """Alternative Keras layers interface for the Keras Resnet V1."""

  def __init__(self,
               batchnorm_training,
               batchnorm_scale=True,
               default_batchnorm_momentum=0.997,
               default_batchnorm_epsilon=1e-5,
               weight_decay=0.0001,
               conv_hyperparams=None,
               min_depth=8,
               depth_multiplier=1):
    """Alternative tf.keras.layers interface, for use by the Keras Resnet V1.

    The class is used by the Keras applications kwargs injection API to
    modify the Resnet V1 Keras application with changes required by
    the Object Detection API.

    Args:
      batchnorm_training: Bool. Assigned to Batch norm layer `training` param
        when constructing `freezable_batch_norm.FreezableBatchNorm` layers.
      batchnorm_scale: If True, uses an explicit `gamma` multiplier to scale
        the activations in the batch normalization layer.
      default_batchnorm_momentum: Float. When 'conv_hyperparams' is None,
        batch norm layers will be constructed using this value as the momentum.
      default_batchnorm_epsilon: Float. When 'conv_hyperparams' is None,
        batch norm layers will be constructed using this value as the epsilon.
      weight_decay: The weight decay to use for regularizing the model.
      conv_hyperparams: A `hyperparams_builder.KerasLayerHyperparams` object
        containing hyperparameters for convolution ops. Optionally set to `None`
        to use default resnet_v1 layer builders.
      min_depth: Minimum number of filters in the convolutional layers.
      depth_multiplier: The depth multiplier to modify the number of filters
        in the convolutional layers.
    """
    self._batchnorm_training = batchnorm_training
    self._batchnorm_scale = batchnorm_scale
    self._default_batchnorm_momentum = default_batchnorm_momentum
    self._default_batchnorm_epsilon = default_batchnorm_epsilon
    self._conv_hyperparams = conv_hyperparams
    self._min_depth = min_depth
    self._depth_multiplier = depth_multiplier
    self.regularizer = tf.keras.regularizers.l2(weight_decay)
    self.initializer = tf.variance_scaling_initializer()

  def _FixedPaddingLayer(self, kernel_size, rate=1):  # pylint: disable=invalid-name
    return tf.keras.layers.Lambda(
        lambda x: _fixed_padding(x, kernel_size, rate))

  def Conv2D(self, filters, kernel_size, **kwargs):  # pylint: disable=invalid-name
    """Builds a Conv2D layer according to the current Object Detection config.

    Overrides the Keras Resnet application's convolutions with ones that
    follow the spec specified by the Object Detection hyperparameters.

    Args:
      filters: The number of filters to use for the convolution.
      kernel_size: The kernel size to specify the height and width of the 2D
        convolution window.
      **kwargs: Keyword args specified by the Keras application for
        constructing the convolution.

    Returns:
      A one-arg callable that will either directly apply a Keras Conv2D layer to
      the input argument, or that will first pad the input then apply a Conv2D
      layer.
    """
    # Apply the minimum depth to the convolution layers.
    filters = max(int(filters * self._depth_multiplier), self._min_depth)

    if self._conv_hyperparams:
      kwargs = self._conv_hyperparams.params(**kwargs)
    else:
      kwargs['kernel_regularizer'] = self.regularizer
      kwargs['kernel_initializer'] = self.initializer

    # Set use_bias as false to keep it consistent with Slim Resnet model.
    kwargs['use_bias'] = False

    kwargs['padding'] = 'same'
    stride = kwargs.get('strides')
    if stride and kernel_size and stride > 1 and kernel_size > 1:
      kwargs['padding'] = 'valid'
      def padded_conv(features):  # pylint: disable=invalid-name
        padded_features = self._FixedPaddingLayer(kernel_size)(features)
        return tf.keras.layers.Conv2D(
            filters, kernel_size, **kwargs)(padded_features)
      return padded_conv
    else:
      return tf.keras.layers.Conv2D(filters, kernel_size, **kwargs)

  def Activation(self, *args, **kwargs):  # pylint: disable=unused-argument,invalid-name
    """Builds an activation layer.

    Overrides the Keras application Activation layer specified by the
    Object Detection configuration.

    Args:
      *args: Ignored,
        required to match the `tf.keras.layers.Activation` interface.
      **kwargs: Only the name is used,
        required to match `tf.keras.layers.Activation` interface.

    Returns:
      An activation layer specified by the Object Detection hyperparameter
      configurations.
    """
    name = kwargs.get('name')
    if self._conv_hyperparams:
      return self._conv_hyperparams.build_activation_layer(name=name)
    else:
      return tf.keras.layers.Lambda(tf.nn.relu, name=name)

  def BatchNormalization(self, **kwargs):  # pylint: disable=invalid-name
    """Builds a normalization layer.

    Overrides the Keras application batch norm with the norm specified by the
    Object Detection configuration.

    Args:
      **kwargs: Only the name is used, all other params ignored.
        Required for matching `layers.BatchNormalization` calls in the Keras
        application.

    Returns:
      A normalization layer specified by the Object Detection hyperparameter
      configurations.
    """
    name = kwargs.get('name')
    if self._conv_hyperparams:
      return self._conv_hyperparams.build_batch_norm(
          training=self._batchnorm_training,
          name=name)
    else:
      kwargs['scale'] = self._batchnorm_scale
      kwargs['epsilon'] = self._default_batchnorm_epsilon
      return freezable_batch_norm.FreezableBatchNorm(
          training=self._batchnorm_training,
          momentum=self._default_batchnorm_momentum,
          **kwargs)

  def Input(self, shape):  # pylint: disable=invalid-name
    """Builds an Input layer.

    Overrides the Keras application Input layer with one that uses a
    tf.placeholder_with_default instead of a tf.placeholder. This is necessary
    to ensure the application works when run on a TPU.

    Args:
      shape: A tuple of integers representing the shape of the input, which
        includes both spatial share and channels, but not the batch size.
        Elements of this tuple can be None; 'None' elements represent dimensions
        where the shape is not known.

    Returns:
      An input layer for the specified shape that internally uses a
      placeholder_with_default.
    """
    default_size = 224
    default_batch_size = 1
    shape = list(shape)
    default_shape = [default_size if dim is None else dim for dim in shape]

    input_tensor = tf.constant(0.0, shape=[default_batch_size] + default_shape)

    placeholder_with_default = tf.placeholder_with_default(
        input=input_tensor, shape=[None] + shape)
    return model_utils.input_layer(shape, placeholder_with_default)

  def MaxPooling2D(self, pool_size, **kwargs):  # pylint: disable=invalid-name
    """Builds a MaxPooling2D layer with default padding as 'SAME'.

    This is specified by the default resnet arg_scope in slim.

    Args:
      pool_size: The pool size specified by the Keras application.
      **kwargs: Ignored, required to match the Keras applications usage.

    Returns:
      A MaxPooling2D layer with default padding as 'SAME'.
    """
    kwargs['padding'] = 'same'
    return tf.keras.layers.MaxPooling2D(pool_size, **kwargs)

  # Add alias as Keras also has it.
  MaxPool2D = MaxPooling2D  # pylint: disable=invalid-name

  def ZeroPadding2D(self, padding, **kwargs):  # pylint: disable=unused-argument,invalid-name
    """Replaces explicit padding in the Keras application with a no-op.

    Args:
      padding: The padding values for image height and width.
      **kwargs: Ignored, required to match the Keras applications usage.

    Returns:
      A no-op identity lambda.
    """
    return lambda x: x

  # Forward all non-overridden methods to the keras layers
  def __getattr__(self, item):
    return getattr(tf.keras.layers, item)


# pylint: disable=invalid-name
def resnet_v1_50(batchnorm_training,
                 batchnorm_scale=True,
                 default_batchnorm_momentum=0.997,
                 default_batchnorm_epsilon=1e-5,
                 weight_decay=0.0001,
                 conv_hyperparams=None,
                 min_depth=8,
                 depth_multiplier=1,
                 **kwargs):
  """Instantiates the Resnet50 architecture, modified for object detection.

  Args:
    batchnorm_training: Bool. Assigned to Batch norm layer `training` param
      when constructing `freezable_batch_norm.FreezableBatchNorm` layers.
    batchnorm_scale: If True, uses an explicit `gamma` multiplier to scale
      the activations in the batch normalization layer.
    default_batchnorm_momentum: Float. When 'conv_hyperparams' is None,
      batch norm layers will be constructed using this value as the momentum.
    default_batchnorm_epsilon: Float. When 'conv_hyperparams' is None,
      batch norm layers will be constructed using this value as the epsilon.
    weight_decay: The weight decay to use for regularizing the model.
    conv_hyperparams: A `hyperparams_builder.KerasLayerHyperparams` object
      containing hyperparameters for convolution ops. Optionally set to `None`
      to use default resnet_v1 layer builders.
    min_depth: Minimum number of filters in the convolutional layers.
    depth_multiplier: The depth multiplier to modify the number of filters
      in the convolutional layers.
    **kwargs: Keyword arguments forwarded directly to the
      `tf.keras.applications.Mobilenet` method that constructs the Keras
      model.

  Returns:
    A Keras ResnetV1-50 model instance.
  """
  layers_override = _LayersOverride(
      batchnorm_training,
      batchnorm_scale=batchnorm_scale,
      default_batchnorm_momentum=default_batchnorm_momentum,
      default_batchnorm_epsilon=default_batchnorm_epsilon,
      conv_hyperparams=conv_hyperparams,
      weight_decay=weight_decay,
      min_depth=min_depth,
      depth_multiplier=depth_multiplier)
  return tf.keras.applications.resnet.ResNet50(
      layers=layers_override, **kwargs)


def resnet_v1_101(batchnorm_training,
                  batchnorm_scale=True,
                  default_batchnorm_momentum=0.997,
                  default_batchnorm_epsilon=1e-5,
                  weight_decay=0.0001,
                  conv_hyperparams=None,
                  min_depth=8,
                  depth_multiplier=1,
                  **kwargs):
  """Instantiates the Resnet50 architecture, modified for object detection.

  Args:
    batchnorm_training: Bool. Assigned to Batch norm layer `training` param
      when constructing `freezable_batch_norm.FreezableBatchNorm` layers.
    batchnorm_scale: If True, uses an explicit `gamma` multiplier to scale
      the activations in the batch normalization layer.
    default_batchnorm_momentum: Float. When 'conv_hyperparams' is None,
      batch norm layers will be constructed using this value as the momentum.
    default_batchnorm_epsilon: Float. When 'conv_hyperparams' is None,
      batch norm layers will be constructed using this value as the epsilon.
    weight_decay: The weight decay to use for regularizing the model.
    conv_hyperparams: A `hyperparams_builder.KerasLayerHyperparams` object
      containing hyperparameters for convolution ops. Optionally set to `None`
      to use default resnet_v1 layer builders.
    min_depth: Minimum number of filters in the convolutional layers.
    depth_multiplier: The depth multiplier to modify the number of filters
      in the convolutional layers.
    **kwargs: Keyword arguments forwarded directly to the
      `tf.keras.applications.Mobilenet` method that constructs the Keras
      model.

  Returns:
    A Keras ResnetV1-101 model instance.
  """
  layers_override = _LayersOverride(
      batchnorm_training,
      batchnorm_scale=batchnorm_scale,
      default_batchnorm_momentum=default_batchnorm_momentum,
      default_batchnorm_epsilon=default_batchnorm_epsilon,
      conv_hyperparams=conv_hyperparams,
      weight_decay=weight_decay,
      min_depth=min_depth,
      depth_multiplier=depth_multiplier)
  return tf.keras.applications.resnet.ResNet101(
      layers=layers_override, **kwargs)


def resnet_v1_152(batchnorm_training,
                  batchnorm_scale=True,
                  default_batchnorm_momentum=0.997,
                  default_batchnorm_epsilon=1e-5,
                  weight_decay=0.0001,
                  conv_hyperparams=None,
                  min_depth=8,
                  depth_multiplier=1,
                  **kwargs):
  """Instantiates the Resnet50 architecture, modified for object detection.

  Args:
    batchnorm_training: Bool. Assigned to Batch norm layer `training` param
      when constructing `freezable_batch_norm.FreezableBatchNorm` layers.
    batchnorm_scale: If True, uses an explicit `gamma` multiplier to scale
      the activations in the batch normalization layer.
    default_batchnorm_momentum: Float. When 'conv_hyperparams' is None,
      batch norm layers will be constructed using this value as the momentum.
    default_batchnorm_epsilon: Float. When 'conv_hyperparams' is None,
      batch norm layers will be constructed using this value as the epsilon.
    weight_decay: The weight decay to use for regularizing the model.
    conv_hyperparams: A `hyperparams_builder.KerasLayerHyperparams` object
      containing hyperparameters for convolution ops. Optionally set to `None`
      to use default resnet_v1 layer builders.
    min_depth: Minimum number of filters in the convolutional layers.
    depth_multiplier: The depth multiplier to modify the number of filters
      in the convolutional layers.
    **kwargs: Keyword arguments forwarded directly to the
      `tf.keras.applications.Mobilenet` method that constructs the Keras
      model.

  Returns:
    A Keras ResnetV1-152 model instance.
  """
  layers_override = _LayersOverride(
      batchnorm_training,
      batchnorm_scale=batchnorm_scale,
      default_batchnorm_momentum=default_batchnorm_momentum,
      default_batchnorm_epsilon=default_batchnorm_epsilon,
      conv_hyperparams=conv_hyperparams,
      weight_decay=weight_decay,
      min_depth=min_depth,
      depth_multiplier=depth_multiplier)
  return tf.keras.applications.resnet.ResNet152(
      layers=layers_override, **kwargs)
# pylint: enable=invalid-name


# The following codes are based on the existing keras ResNet model pattern:
# google3/third_party/tensorflow/python/keras/applications/resnet.py
def block_basic(x,
                filters,
                kernel_size=3,
                stride=1,
                conv_shortcut=False,
                name=None):
  """A residual block for ResNet18/34.

  Args:
      x: input tensor.
      filters: integer, filters of the bottleneck layer.
      kernel_size: default 3, kernel size of the bottleneck layer.
      stride: default 1, stride of the first layer.
      conv_shortcut: default False, use convolution shortcut if True, otherwise
        identity shortcut.
      name: string, block label.

  Returns:
    Output tensor for the residual block.
  """
  layers = tf.keras.layers
  bn_axis = 3 if tf.keras.backend.image_data_format() == 'channels_last' else 1

  preact = layers.BatchNormalization(
      axis=bn_axis, epsilon=1.001e-5, name=name + '_preact_bn')(
          x)
  preact = layers.Activation('relu', name=name + '_preact_relu')(preact)

  if conv_shortcut:
    shortcut = layers.Conv2D(
        filters, 1, strides=1, name=name + '_0_conv')(
            preact)
  else:
    shortcut = layers.MaxPooling2D(1, strides=stride)(x) if stride > 1 else x

  x = layers.ZeroPadding2D(
      padding=((1, 1), (1, 1)), name=name + '_1_pad')(
          preact)
  x = layers.Conv2D(
      filters, kernel_size, strides=1, use_bias=False, name=name + '_1_conv')(
          x)
  x = layers.BatchNormalization(
      axis=bn_axis, epsilon=1.001e-5, name=name + '_1_bn')(
          x)
  x = layers.Activation('relu', name=name + '_1_relu')(x)

  x = layers.ZeroPadding2D(padding=((1, 1), (1, 1)), name=name + '_2_pad')(x)
  x = layers.Conv2D(
      filters,
      kernel_size,
      strides=stride,
      use_bias=False,
      name=name + '_2_conv')(
          x)
  x = layers.BatchNormalization(
      axis=bn_axis, epsilon=1.001e-5, name=name + '_2_bn')(
          x)
  x = layers.Activation('relu', name=name + '_2_relu')(x)
  x = layers.Add(name=name + '_out')([shortcut, x])
  return x


def stack_basic(x, filters, blocks, stride1=2, name=None):
  """A set of stacked residual blocks for ResNet18/34.

  Args:
      x: input tensor.
      filters: integer, filters of the bottleneck layer in a block.
      blocks: integer, blocks in the stacked blocks.
      stride1: default 2, stride of the first layer in the first block.
      name: string, stack label.

  Returns:
      Output tensor for the stacked blocks.
  """
  x = block_basic(x, filters, conv_shortcut=True, name=name + '_block1')
  for i in range(2, blocks):
    x = block_basic(x, filters, name=name + '_block' + str(i))
  x = block_basic(
      x, filters, stride=stride1, name=name + '_block' + str(blocks))
  return x


def resnet_v1_18(include_top=True,
                 weights='imagenet',
                 input_tensor=None,
                 input_shape=None,
                 pooling=None,
                 classes=1000,
                 classifier_activation='softmax'):
  """Instantiates the ResNet18 architecture."""

  def stack_fn(x):
    x = stack_basic(x, 64, 2, stride1=1, name='conv2')
    x = stack_basic(x, 128, 2, name='conv3')
    x = stack_basic(x, 256, 2, name='conv4')
    return stack_basic(x, 512, 2, name='conv5')

  return resnet.ResNet(
      stack_fn,
      True,
      True,
      'resnet18',
      include_top,
      weights,
      input_tensor,
      input_shape,
      pooling,
      classes,
      classifier_activation=classifier_activation)


def resnet_v1_34(include_top=True,
                 weights='imagenet',
                 input_tensor=None,
                 input_shape=None,
                 pooling=None,
                 classes=1000,
                 classifier_activation='softmax'):
  """Instantiates the ResNet34 architecture."""

  def stack_fn(x):
    x = stack_basic(x, 64, 3, stride1=1, name='conv2')
    x = stack_basic(x, 128, 4, name='conv3')
    x = stack_basic(x, 256, 6, name='conv4')
    return stack_basic(x, 512, 3, name='conv5')

  return resnet.ResNet(
      stack_fn,
      True,
      True,
      'resnet34',
      include_top,
      weights,
      input_tensor,
      input_shape,
      pooling,
      classes,
      classifier_activation=classifier_activation)



# Copyright (c) Facebook, Inc. and its affiliates.

import unittest
import torch
from torch import nn

from detectron2.layers import ASPP, DepthwiseSeparableConv2d, FrozenBatchNorm2d
from detectron2.modeling.backbone.resnet import BasicStem, ResNet


"""
Test for misc layers.
"""


class TestBlocks(unittest.TestCase):
    def test_separable_conv(self):
        DepthwiseSeparableConv2d(3, 10, norm1="BN", activation1=nn.PReLU())

    def test_aspp(self):
        m = ASPP(3, 10, [2, 3, 4], norm="", activation=nn.PReLU())
        self.assertIsNot(m.convs[0].activation.weight, m.convs[1].activation.weight)
        self.assertIsNot(m.convs[0].activation.weight, m.project.activation.weight)

    @unittest.skipIf(not torch.cuda.is_available(), "CUDA not available")
    def test_frozen_batchnorm_fp16(self):
        from torch.cuda.amp import autocast

        C = 10
        input = torch.rand(1, C, 10, 10).cuda()
        m = FrozenBatchNorm2d(C).cuda()
        with autocast():
            output = m(input.half())
        self.assertEqual(output.dtype, torch.float16)

        # requires_grad triggers a different codepath
        input.requires_grad_()
        with autocast():
            output = m(input.half())
        self.assertEqual(output.dtype, torch.float16)

    def test_resnet_unused_stages(self):
        resnet = ResNet(BasicStem(), ResNet.make_default_stages(18), out_features=["res2"])
        self.assertTrue(hasattr(resnet, "res2"))
        self.assertFalse(hasattr(resnet, "res3"))
        self.assertFalse(hasattr(resnet, "res5"))

        resnet = ResNet(BasicStem(), ResNet.make_default_stages(18), out_features=["res2", "res5"])
        self.assertTrue(hasattr(resnet, "res2"))
        self.assertTrue(hasattr(resnet, "res4"))
        self.assertTrue(hasattr(resnet, "res5"))



# Lint as: python2, python3
# Copyright 2020 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""SSD MobilenetV2 NAS-FPN Feature Extractor."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import collections
import functools
from six.moves import range
import tensorflow.compat.v1 as tf

import tf_slim as slim
from object_detection.meta_architectures import ssd_meta_arch
from object_detection.utils import ops
from object_detection.utils import shape_utils
from nets.mobilenet import mobilenet
from nets.mobilenet import mobilenet_v2


Block = collections.namedtuple(
    'Block', ['inputs', 'output_level', 'kernel_size', 'expansion_size'])

_MNASFPN_CELL_CONFIG = [
    Block(inputs=(1, 2), output_level=4, kernel_size=3, expansion_size=256),
    Block(inputs=(0, 4), output_level=3, kernel_size=3, expansion_size=128),
    Block(inputs=(5, 4), output_level=4, kernel_size=3, expansion_size=128),
    Block(inputs=(4, 3), output_level=5, kernel_size=5, expansion_size=128),
    Block(inputs=(4, 3), output_level=6, kernel_size=3, expansion_size=96),
]

MNASFPN_DEF = dict(
    feature_levels=[3, 4, 5, 6],
    spec=[_MNASFPN_CELL_CONFIG] * 4,
)


def _maybe_pad(feature, use_explicit_padding, kernel_size=3):
  return ops.fixed_padding(feature,
                           kernel_size) if use_explicit_padding else feature


# Wrapper around mobilenet.depth_multiplier
def _apply_multiplier(d, multiplier, min_depth):
  p = {'num_outputs': d}
  mobilenet.depth_multiplier(
      p, multiplier=multiplier, divisible_by=8, min_depth=min_depth)
  return p['num_outputs']


def _apply_size_dependent_ordering(input_feature, feature_level, block_level,
                                   expansion_size, use_explicit_padding,
                                   use_native_resize_op):
  """Applies Size-Dependent-Ordering when resizing feature maps.

     See https://arxiv.org/abs/1912.01106

  Args:
    input_feature: input feature map to be resized.
    feature_level: the level of the input feature.
    block_level: the desired output level for the block.
    expansion_size: the expansion size for the block.
    use_explicit_padding: Whether to use explicit padding.
    use_native_resize_op: Whether to use native resize op.

  Returns:
    A transformed feature at the desired resolution and expansion size.
  """
  padding = 'VALID' if use_explicit_padding else 'SAME'
  if feature_level >= block_level:  # Perform 1x1 then upsampling.
    node = slim.conv2d(
        input_feature,
        expansion_size, [1, 1],
        activation_fn=None,
        normalizer_fn=slim.batch_norm,
        padding=padding,
        scope='Conv1x1')
    if feature_level == block_level:
      return node
    scale = 2**(feature_level - block_level)
    if use_native_resize_op:
      input_shape = shape_utils.combined_static_and_dynamic_shape(node)
      node = tf.image.resize_nearest_neighbor(
          node, [input_shape[1] * scale, input_shape[2] * scale])
    else:
      node = ops.nearest_neighbor_upsampling(node, scale=scale)
  else:  # Perform downsampling then 1x1.
    stride = 2**(block_level - feature_level)
    node = slim.max_pool2d(
        _maybe_pad(input_feature, use_explicit_padding), [3, 3],
        stride=[stride, stride],
        padding=padding,
        scope='Downsample')
    node = slim.conv2d(
        node,
        expansion_size, [1, 1],
        activation_fn=None,
        normalizer_fn=slim.batch_norm,
        padding=padding,
        scope='Conv1x1')
  return node


def _mnasfpn_cell(feature_maps,
                  feature_levels,
                  cell_spec,
                  output_channel=48,
                  use_explicit_padding=False,
                  use_native_resize_op=False,
                  multiplier_func=None):
  """Create a MnasFPN cell.

  Args:
    feature_maps: input feature maps.
    feature_levels: levels of the feature maps.
    cell_spec: A list of Block configs.
    output_channel: Number of features for the input, output and intermediate
      feature maps.
    use_explicit_padding: Whether to use explicit padding.
    use_native_resize_op: Whether to use native resize op.
    multiplier_func: Depth-multiplier function. If None, use identity function.

  Returns:
    A transformed list of feature maps at the same resolutions as the inputs.
  """
  # This is the level where multipliers are realized.
  if multiplier_func is None:
    multiplier_func = lambda x: x
  num_outputs = len(feature_maps)
  cell_features = list(feature_maps)
  cell_levels = list(feature_levels)
  padding = 'VALID' if use_explicit_padding else 'SAME'
  for bi, block in enumerate(cell_spec):
    with tf.variable_scope('block_{}'.format(bi)):
      block_level = block.output_level
      intermediate_feature = None
      for i, inp in enumerate(block.inputs):
        with tf.variable_scope('input_{}'.format(i)):
          input_level = cell_levels[inp]
          node = _apply_size_dependent_ordering(
              cell_features[inp], input_level, block_level,
              multiplier_func(block.expansion_size), use_explicit_padding,
              use_native_resize_op)
        # Add features incrementally to avoid producing AddN, which doesn't
        # play well with TfLite.
        if intermediate_feature is None:
          intermediate_feature = node
        else:
          intermediate_feature += node
      node = tf.nn.relu6(intermediate_feature)
      node = slim.separable_conv2d(
          _maybe_pad(node, use_explicit_padding, block.kernel_size),
          multiplier_func(output_channel),
          block.kernel_size,
          activation_fn=None,
          normalizer_fn=slim.batch_norm,
          padding=padding,
          scope='SepConv')
    cell_features.append(node)
    cell_levels.append(block_level)

  # Cell-wide residuals.
  out_idx = range(len(cell_features) - num_outputs, len(cell_features))
  for in_i, out_i in enumerate(out_idx):
    if cell_features[out_i].shape.as_list(
    ) == cell_features[in_i].shape.as_list():
      cell_features[out_i] += cell_features[in_i]

  return cell_features[-num_outputs:]


def mnasfpn(feature_maps,
            head_def,
            output_channel=48,
            use_explicit_padding=False,
            use_native_resize_op=False,
            multiplier_func=None):
  """Create the MnasFPN head given head_def."""
  features = feature_maps
  for ci, cell_spec in enumerate(head_def['spec']):
    with tf.variable_scope('cell_{}'.format(ci)):
      features = _mnasfpn_cell(features, head_def['feature_levels'], cell_spec,
                               output_channel, use_explicit_padding,
                               use_native_resize_op, multiplier_func)
  return features


def training_scope(l2_weight_decay=1e-4, is_training=None):
  """Arg scope for training MnasFPN."""
  with slim.arg_scope(
      [slim.conv2d],
      weights_initializer=tf.initializers.he_normal(),
      weights_regularizer=slim.l2_regularizer(l2_weight_decay)), \
      slim.arg_scope(
          [slim.separable_conv2d],
          weights_initializer=tf.initializers.truncated_normal(
              stddev=0.536),  # He_normal for 3x3 depthwise kernel.
          weights_regularizer=slim.l2_regularizer(l2_weight_decay)), \
      slim.arg_scope([slim.batch_norm],
                     is_training=is_training,
                     epsilon=0.01,
                     decay=0.99,
                     center=True,
                     scale=True) as s:
    return s


class SSDMobileNetV2MnasFPNFeatureExtractor(ssd_meta_arch.SSDFeatureExtractor):
  """SSD Feature Extractor using MobilenetV2 MnasFPN features."""

  def __init__(self,
               is_training,
               depth_multiplier,
               min_depth,
               pad_to_multiple,
               conv_hyperparams_fn,
               fpn_min_level=3,
               fpn_max_level=6,
               additional_layer_depth=48,
               head_def=None,
               reuse_weights=None,
               use_explicit_padding=False,
               use_depthwise=False,
               use_native_resize_op=False,
               override_base_feature_extractor_hyperparams=False,
               data_format='channels_last'):
    """SSD MnasFPN feature extractor based on Mobilenet v2 architecture.

    See https://arxiv.org/abs/1912.01106

    Args:
      is_training: whether the network is in training mode.
      depth_multiplier: float depth multiplier for feature extractor.
      min_depth: minimum feature extractor depth.
      pad_to_multiple: the nearest multiple to zero pad the input height and
        width dimensions to.
      conv_hyperparams_fn: A function to construct tf slim arg_scope for conv2d
        and separable_conv2d ops in the layers that are added on top of the base
        feature extractor.
      fpn_min_level: the highest resolution feature map to use in MnasFPN.
        Currently the only valid value is 3.
      fpn_max_level: the smallest resolution feature map to construct or use in
        MnasFPN. Currentl the only valid value is 6.
      additional_layer_depth: additional feature map layer channel depth for
        NAS-FPN.
      head_def: A dictionary specifying the MnasFPN head architecture. Default
        uses MNASFPN_DEF.
      reuse_weights: whether to reuse variables. Default is None.
      use_explicit_padding: Whether to use explicit padding when extracting
        features. Default is False.
      use_depthwise: Whether to use depthwise convolutions. Default is False.
      use_native_resize_op: Whether to use native resize op. Default is False.
      override_base_feature_extractor_hyperparams: Whether to override
        hyperparameters of the base feature extractor with the one from
        `conv_hyperparams_fn`.
      data_format: The ordering of the dimensions in the inputs, The valid
        values are {'channels_first', 'channels_last').
    """
    super(SSDMobileNetV2MnasFPNFeatureExtractor, self).__init__(
        is_training=is_training,
        depth_multiplier=depth_multiplier,
        min_depth=min_depth,
        pad_to_multiple=pad_to_multiple,
        conv_hyperparams_fn=conv_hyperparams_fn,
        reuse_weights=reuse_weights,
        use_explicit_padding=use_explicit_padding,
        use_depthwise=use_depthwise,
        override_base_feature_extractor_hyperparams=(
            override_base_feature_extractor_hyperparams))
    if fpn_min_level != 3 or fpn_max_level != 6:
      raise ValueError('Min and max levels of MnasFPN must be 3 and 6 for now.')
    self._fpn_min_level = fpn_min_level
    self._fpn_max_level = fpn_max_level
    self._fpn_layer_depth = additional_layer_depth
    self._head_def = head_def if head_def else MNASFPN_DEF
    self._data_format = data_format
    self._use_native_resize_op = use_native_resize_op

  def preprocess(self, resized_inputs):
    """SSD preprocessing.

    Maps pixel values to the range [-1, 1].

    Args:
      resized_inputs: a [batch, height, width, channels] float tensor
        representing a batch of images.

    Returns:
      preprocessed_inputs: a [batch, height, width, channels] float tensor
        representing a batch of images.
    """
    return (2.0 / 255.0) * resized_inputs - 1.0

  def _verify_config(self, inputs):
    """Verify that MnasFPN config and its inputs."""
    num_inputs = len(inputs)
    assert len(self._head_def['feature_levels']) == num_inputs

    base_width = inputs[0].shape.as_list(
    )[1] * 2**self._head_def['feature_levels'][0]
    for i in range(1, num_inputs):
      width = inputs[i].shape.as_list()[1]
      level = self._head_def['feature_levels'][i]
      expected_width = base_width // 2**level
      if width != expected_width:
        raise ValueError(
            'Resolution of input {} does not match its level {}.'.format(
                i, level))

    for cell_spec in self._head_def['spec']:
      # The last K nodes in a cell are the inputs to the next cell. Assert that
      # their feature maps are at the right level.
      for i in range(num_inputs):
        if cell_spec[-num_inputs +
                     i].output_level != self._head_def['feature_levels'][i]:
          raise ValueError(
              'Mismatch between node level {} and desired output level {}.'
              .format(cell_spec[-num_inputs + i].output_level,
                      self._head_def['feature_levels'][i]))
      # Assert that each block only uses precending blocks.
      for bi, block_spec in enumerate(cell_spec):
        for inp in block_spec.inputs:
          if inp >= bi + num_inputs:
            raise ValueError(
                'Block {} is trying to access uncreated block {}.'.format(
                    bi, inp))

  def extract_features(self, preprocessed_inputs):
    """Extract features from preprocessed inputs.

    Args:
      preprocessed_inputs: a [batch, height, width, channels] float tensor
        representing a batch of images.

    Returns:
      feature_maps: a list of tensors where the ith tensor has shape
        [batch, height_i, width_i, depth_i]
    """
    preprocessed_inputs = shape_utils.check_min_image_dim(
        33, preprocessed_inputs)
    with tf.variable_scope('MobilenetV2', reuse=self._reuse_weights) as scope:
      with slim.arg_scope(
          mobilenet_v2.training_scope(is_training=None, bn_decay=0.99)), \
          slim.arg_scope(
              [mobilenet.depth_multiplier], min_depth=self._min_depth):
        with slim.arg_scope(
            training_scope(l2_weight_decay=4e-5,
                           is_training=self._is_training)):

          _, image_features = mobilenet_v2.mobilenet_base(
              ops.pad_to_multiple(preprocessed_inputs, self._pad_to_multiple),
              final_endpoint='layer_18',
              depth_multiplier=self._depth_multiplier,
              use_explicit_padding=self._use_explicit_padding,
              scope=scope)

    multiplier_func = functools.partial(
        _apply_multiplier,
        multiplier=self._depth_multiplier,
        min_depth=self._min_depth)
    with tf.variable_scope('MnasFPN', reuse=self._reuse_weights):
      with slim.arg_scope(
          training_scope(l2_weight_decay=1e-4, is_training=self._is_training)):
        # Create C6 by downsampling C5.
        c6 = slim.max_pool2d(
            _maybe_pad(image_features['layer_18'], self._use_explicit_padding),
            [3, 3],
            stride=[2, 2],
            padding='VALID' if self._use_explicit_padding else 'SAME',
            scope='C6_downsample')
        c6 = slim.conv2d(
            c6,
            multiplier_func(self._fpn_layer_depth),
            [1, 1],
            activation_fn=tf.identity,
            normalizer_fn=slim.batch_norm,
            weights_regularizer=None,  # this 1x1 has no kernel regularizer.
            padding='VALID',
            scope='C6_Conv1x1')
        image_features['C6'] = tf.identity(c6)  # Needed for quantization.
        for k in sorted(image_features.keys()):
          tf.logging.error('{}: {}'.format(k, image_features[k]))

        mnasfpn_inputs = [
            image_features['layer_7'],  # C3
            image_features['layer_14'],  # C4
            image_features['layer_18'],  # C5
            image_features['C6']  # C6
        ]
        self._verify_config(mnasfpn_inputs)
        feature_maps = mnasfpn(
            mnasfpn_inputs,
            head_def=self._head_def,
            output_channel=self._fpn_layer_depth,
            use_explicit_padding=self._use_explicit_padding,
            use_native_resize_op=self._use_native_resize_op,
            multiplier_func=multiplier_func)
    return feature_maps



# YOLOv5 🚀 by Ultralytics, GPL-3.0 license
"""
utils/initialization
"""


def notebook_init(verbose=True):
    # Check system software and hardware
    print('Checking setup...')

    import os
    import shutil

    from utils.general import check_requirements, emojis, is_colab
    from utils.torch_utils import select_device  # imports

    check_requirements(('psutil', 'IPython'))
    import psutil
    from IPython import display  # to display images and clear console output

    if is_colab():
        shutil.rmtree('/content/sample_data', ignore_errors=True)  # remove colab /sample_data directory

    # System info
    if verbose:
        gb = 1 << 30  # bytes to GiB (1024 ** 3)
        ram = psutil.virtual_memory().total
        total, used, free = shutil.disk_usage("/")
        display.clear_output()
        s = f'({os.cpu_count()} CPUs, {ram / gb:.1f} GB RAM, {(total - free) / gb:.1f}/{total / gb:.1f} GB disk)'
    else:
        s = ''

    select_device(newline=False)
    print(emojis(f'Setup complete ✅ {s}'))
    return display



# Copyright 2017 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""Tests for object_detection.core.box_list_ops."""
import numpy as np
import tensorflow.compat.v1 as tf

from object_detection.core import box_list
from object_detection.core import box_list_ops
from object_detection.utils import test_case


class BoxListOpsTest(test_case.TestCase):
  """Tests for common bounding box operations."""

  def test_area(self):
    def graph_fn():
      corners = tf.constant([[0.0, 0.0, 10.0, 20.0], [1.0, 2.0, 3.0, 4.0]])
      boxes = box_list.BoxList(corners)
      areas = box_list_ops.area(boxes)
      return areas
    areas_out = self.execute(graph_fn, [])
    exp_output = [200.0, 4.0]
    self.assertAllClose(areas_out, exp_output)

  def test_height_width(self):
    def graph_fn():
      corners = tf.constant([[0.0, 0.0, 10.0, 20.0], [1.0, 2.0, 3.0, 4.0]])
      boxes = box_list.BoxList(corners)
      return box_list_ops.height_width(boxes)
    heights_out, widths_out = self.execute(graph_fn, [])
    exp_output_heights = [10., 2.]
    exp_output_widths = [20., 2.]
    self.assertAllClose(heights_out, exp_output_heights)
    self.assertAllClose(widths_out, exp_output_widths)

  def test_scale(self):
    def graph_fn():
      corners = tf.constant([[0, 0, 100, 200], [50, 120, 100, 140]],
                            dtype=tf.float32)
      boxes = box_list.BoxList(corners)
      boxes.add_field('extra_data', tf.constant([[1], [2]]))

      y_scale = tf.constant(1.0/100)
      x_scale = tf.constant(1.0/200)
      scaled_boxes = box_list_ops.scale(boxes, y_scale, x_scale)
      return scaled_boxes.get(), scaled_boxes.get_field('extra_data')
    scaled_corners_out, extra_data_out = self.execute(graph_fn, [])
    exp_output = [[0, 0, 1, 1], [0.5, 0.6, 1.0, 0.7]]
    self.assertAllClose(scaled_corners_out, exp_output)
    self.assertAllEqual(extra_data_out, [[1], [2]])

  def test_scale_height_width(self):
    def graph_fn():
      corners = tf.constant([[-10, -20, 10, 20], [0, 100, 100, 200]],
                            dtype=tf.float32)
      boxes = box_list.BoxList(corners)
      boxes.add_field('extra_data', tf.constant([[1], [2]]))

      y_scale = tf.constant(2.)
      x_scale = tf.constant(0.5)
      scaled_boxes = box_list_ops.scale_height_width(boxes, y_scale, x_scale)
      return scaled_boxes.get(), scaled_boxes.get_field('extra_data')
    exp_output = [
        [-20., -10, 20., 10],
        [-50., 125, 150., 175.]]
    scaled_corners_out, extra_data_out = self.execute(graph_fn, [])
    self.assertAllClose(scaled_corners_out, exp_output)
    self.assertAllEqual(extra_data_out, [[1], [2]])

  def test_clip_to_window_filter_boxes_which_fall_outside_the_window(
      self):
    def graph_fn():
      window = tf.constant([0, 0, 9, 14], tf.float32)
      corners = tf.constant([[5.0, 5.0, 6.0, 6.0],
                             [-1.0, -2.0, 4.0, 5.0],
                             [2.0, 3.0, 5.0, 9.0],
                             [0.0, 0.0, 9.0, 14.0],
                             [-100.0, -100.0, 300.0, 600.0],
                             [-10.0, -10.0, -9.0, -9.0]])
      boxes = box_list.BoxList(corners)
      boxes.add_field('extra_data', tf.constant([[1], [2], [3], [4], [5], [6]]))
      pruned = box_list_ops.clip_to_window(
          boxes, window, filter_nonoverlapping=True)
      return pruned.get(), pruned.get_field('extra_data')
    exp_output = [[5.0, 5.0, 6.0, 6.0], [0.0, 0.0, 4.0, 5.0],
                  [2.0, 3.0, 5.0, 9.0], [0.0, 0.0, 9.0, 14.0],
                  [0.0, 0.0, 9.0, 14.0]]
    pruned_output, extra_data_out = self.execute_cpu(graph_fn, [])
    self.assertAllClose(pruned_output, exp_output)
    self.assertAllEqual(extra_data_out, [[1], [2], [3], [4], [5]])

  def test_clip_to_window_without_filtering_boxes_which_fall_outside_the_window(
      self):
    def graph_fn():
      window = tf.constant([0, 0, 9, 14], tf.float32)
      corners = tf.constant([[5.0, 5.0, 6.0, 6.0],
                             [-1.0, -2.0, 4.0, 5.0],
                             [2.0, 3.0, 5.0, 9.0],
                             [0.0, 0.0, 9.0, 14.0],
                             [-100.0, -100.0, 300.0, 600.0],
                             [-10.0, -10.0, -9.0, -9.0]])
      boxes = box_list.BoxList(corners)
      boxes.add_field('extra_data', tf.constant([[1], [2], [3], [4], [5], [6]]))
      pruned = box_list_ops.clip_to_window(
          boxes, window, filter_nonoverlapping=False)
      return pruned.get(), pruned.get_field('extra_data')
    pruned_output, extra_data_out = self.execute(graph_fn, [])
    exp_output = [[5.0, 5.0, 6.0, 6.0], [0.0, 0.0, 4.0, 5.0],
                  [2.0, 3.0, 5.0, 9.0], [0.0, 0.0, 9.0, 14.0],
                  [0.0, 0.0, 9.0, 14.0], [0.0, 0.0, 0.0, 0.0]]
    self.assertAllClose(pruned_output, exp_output)
    self.assertAllEqual(extra_data_out, [[1], [2], [3], [4], [5], [6]])

  def test_prune_outside_window_filters_boxes_which_fall_outside_the_window(
      self):
    def graph_fn():
      window = tf.constant([0, 0, 9, 14], tf.float32)
      corners = tf.constant([[5.0, 5.0, 6.0, 6.0],
                             [-1.0, -2.0, 4.0, 5.0],
                             [2.0, 3.0, 5.0, 9.0],
                             [0.0, 0.0, 9.0, 14.0],
                             [-10.0, -10.0, -9.0, -9.0],
                             [-100.0, -100.0, 300.0, 600.0]])
      boxes = box_list.BoxList(corners)
      boxes.add_field('extra_data', tf.constant([[1], [2], [3], [4], [5], [6]]))
      pruned, keep_indices = box_list_ops.prune_outside_window(boxes, window)
      return pruned.get(), pruned.get_field('extra_data'), keep_indices
    pruned_output, extra_data_out, keep_indices_out = self.execute_cpu(graph_fn,
                                                                       [])
    exp_output = [[5.0, 5.0, 6.0, 6.0],
                  [2.0, 3.0, 5.0, 9.0],
                  [0.0, 0.0, 9.0, 14.0]]
    self.assertAllClose(pruned_output, exp_output)
    self.assertAllEqual(keep_indices_out, [0, 2, 3])
    self.assertAllEqual(extra_data_out, [[1], [3], [4]])

  def test_prune_completely_outside_window(self):
    def graph_fn():
      window = tf.constant([0, 0, 9, 14], tf.float32)
      corners = tf.constant([[5.0, 5.0, 6.0, 6.0],
                             [-1.0, -2.0, 4.0, 5.0],
                             [2.0, 3.0, 5.0, 9.0],
                             [0.0, 0.0, 9.0, 14.0],
                             [-10.0, -10.0, -9.0, -9.0],
                             [-100.0, -100.0, 300.0, 600.0]])
      boxes = box_list.BoxList(corners)
      boxes.add_field('extra_data', tf.constant([[1], [2], [3], [4], [5], [6]]))
      pruned, keep_indices = box_list_ops.prune_completely_outside_window(
          boxes, window)
      return pruned.get(), pruned.get_field('extra_data'), keep_indices
    pruned_output, extra_data_out, keep_indices_out = self.execute(graph_fn, [])
    exp_output = [[5.0, 5.0, 6.0, 6.0],
                  [-1.0, -2.0, 4.0, 5.0],
                  [2.0, 3.0, 5.0, 9.0],
                  [0.0, 0.0, 9.0, 14.0],
                  [-100.0, -100.0, 300.0, 600.0]]
    self.assertAllClose(pruned_output, exp_output)
    self.assertAllEqual(keep_indices_out, [0, 1, 2, 3, 5])
    self.assertAllEqual(extra_data_out, [[1], [2], [3], [4], [6]])

  def test_prune_completely_outside_window_with_empty_boxlist(self):
    def graph_fn():
      window = tf.constant([0, 0, 9, 14], tf.float32)
      corners = tf.zeros(shape=[0, 4], dtype=tf.float32)
      boxes = box_list.BoxList(corners)
      boxes.add_field('extra_data', tf.zeros(shape=[0], dtype=tf.int32))
      pruned, keep_indices = box_list_ops.prune_completely_outside_window(
          boxes, window)
      pruned_boxes = pruned.get()
      extra = pruned.get_field('extra_data')
      return pruned_boxes, extra, keep_indices

    pruned_boxes_out, extra_out, keep_indices_out = self.execute(graph_fn, [])
    exp_pruned_boxes = np.zeros(shape=[0, 4], dtype=np.float32)
    exp_extra = np.zeros(shape=[0], dtype=np.int32)
    self.assertAllClose(exp_pruned_boxes, pruned_boxes_out)
    self.assertAllEqual([], keep_indices_out)
    self.assertAllEqual(exp_extra, extra_out)

  def test_intersection(self):
    def graph_fn():
      corners1 = tf.constant([[4.0, 3.0, 7.0, 5.0], [5.0, 6.0, 10.0, 7.0]])
      corners2 = tf.constant([[3.0, 4.0, 6.0, 8.0], [14.0, 14.0, 15.0, 15.0],
                              [0.0, 0.0, 20.0, 20.0]])
      boxes1 = box_list.BoxList(corners1)
      boxes2 = box_list.BoxList(corners2)
      intersect = box_list_ops.intersection(boxes1, boxes2)
      return intersect
    exp_output = [[2.0, 0.0, 6.0], [1.0, 0.0, 5.0]]
    intersect_out = self.execute(graph_fn, [])
    self.assertAllClose(intersect_out, exp_output)

  def test_matched_intersection(self):
    def graph_fn():
      corners1 = tf.constant([[4.0, 3.0, 7.0, 5.0], [5.0, 6.0, 10.0, 7.0]])
      corners2 = tf.constant([[3.0, 4.0, 6.0, 8.0], [14.0, 14.0, 15.0, 15.0]])
      boxes1 = box_list.BoxList(corners1)
      boxes2 = box_list.BoxList(corners2)
      intersect = box_list_ops.matched_intersection(boxes1, boxes2)
      return intersect
    exp_output = [2.0, 0.0]
    intersect_out = self.execute(graph_fn, [])
    self.assertAllClose(intersect_out, exp_output)

  def test_iou(self):
    def graph_fn():
      corners1 = tf.constant([[4.0, 3.0, 7.0, 5.0], [5.0, 6.0, 10.0, 7.0]])
      corners1 = tf.constant([[4.0, 3.0, 7.0, 5.0], [5.0, 6.0, 10.0, 7.0]])
      corners2 = tf.constant([[3.0, 4.0, 6.0, 8.0], [14.0, 14.0, 15.0, 15.0],
                              [0.0, 0.0, 20.0, 20.0]])
      boxes1 = box_list.BoxList(corners1)
      boxes2 = box_list.BoxList(corners2)
      iou = box_list_ops.iou(boxes1, boxes2)
      return iou
    exp_output = [[2.0 / 16.0, 0, 6.0 / 400.0], [1.0 / 16.0, 0.0, 5.0 / 400.0]]
    iou_output = self.execute(graph_fn, [])
    self.assertAllClose(iou_output, exp_output)

  def test_l1(self):
    def graph_fn():
      corners1 = tf.constant([[4.0, 3.0, 7.0, 5.0], [5.0, 6.0, 10.0, 7.0]])
      corners2 = tf.constant([[3.0, 4.0, 6.0, 8.0], [14.0, 14.0, 15.0, 15.0],
                              [0.0, 0.0, 20.0, 20.0]])
      boxes1 = box_list.BoxList(corners1)
      boxes2 = box_list.BoxList(corners2)
      l1 = box_list_ops.l1(boxes1, boxes2)
      return l1
    exp_output = [[5.0, 22.5, 45.5], [8.5, 19.0, 40.0]]
    l1_output = self.execute(graph_fn, [])
    self.assertAllClose(l1_output, exp_output)

  def test_giou(self):
    def graph_fn():
      corners1 = tf.constant([[5.0, 7.0, 7.0, 9.0]])
      corners2 = tf.constant([[5.0, 7.0, 7.0, 9.0], [5.0, 11.0, 7.0, 13.0]])
      boxes1 = box_list.BoxList(corners1)
      boxes2 = box_list.BoxList(corners2)
      giou = box_list_ops.giou(boxes1, boxes2)
      return giou
    exp_output = [[1.0, -1.0 / 3.0]]
    giou_output = self.execute(graph_fn, [])
    self.assertAllClose(giou_output, exp_output)

  def test_matched_iou(self):
    def graph_fn():
      corners1 = tf.constant([[4.0, 3.0, 7.0, 5.0], [5.0, 6.0, 10.0, 7.0]])
      corners2 = tf.constant([[3.0, 4.0, 6.0, 8.0], [14.0, 14.0, 15.0, 15.0]])
      boxes1 = box_list.BoxList(corners1)
      boxes2 = box_list.BoxList(corners2)
      iou = box_list_ops.matched_iou(boxes1, boxes2)
      return iou
    exp_output = [2.0 / 16.0, 0]
    iou_output = self.execute(graph_fn, [])
    self.assertAllClose(iou_output, exp_output)

  def test_iouworks_on_empty_inputs(self):
    def graph_fn():
      corners1 = tf.constant([[4.0, 3.0, 7.0, 5.0], [5.0, 6.0, 10.0, 7.0]])
      corners2 = tf.constant([[3.0, 4.0, 6.0, 8.0], [14.0, 14.0, 15.0, 15.0],
                              [0.0, 0.0, 20.0, 20.0]])
      boxes1 = box_list.BoxList(corners1)
      boxes2 = box_list.BoxList(corners2)
      boxes_empty = box_list.BoxList(tf.zeros((0, 4)))
      iou_empty_1 = box_list_ops.iou(boxes1, boxes_empty)
      iou_empty_2 = box_list_ops.iou(boxes_empty, boxes2)
      iou_empty_3 = box_list_ops.iou(boxes_empty, boxes_empty)
      return iou_empty_1, iou_empty_2, iou_empty_3
    iou_output_1, iou_output_2, iou_output_3 = self.execute(graph_fn, [])
    self.assertAllEqual(iou_output_1.shape, (2, 0))
    self.assertAllEqual(iou_output_2.shape, (0, 3))
    self.assertAllEqual(iou_output_3.shape, (0, 0))

  def test_ioa(self):
    def graph_fn():
      corners1 = tf.constant([[4.0, 3.0, 7.0, 5.0], [5.0, 6.0, 10.0, 7.0]])
      corners2 = tf.constant([[3.0, 4.0, 6.0, 8.0], [14.0, 14.0, 15.0, 15.0],
                              [0.0, 0.0, 20.0, 20.0]])
      boxes1 = box_list.BoxList(corners1)
      boxes2 = box_list.BoxList(corners2)
      ioa_1 = box_list_ops.ioa(boxes1, boxes2)
      ioa_2 = box_list_ops.ioa(boxes2, boxes1)
      return ioa_1, ioa_2
    exp_output_1 = [[2.0 / 12.0, 0, 6.0 / 400.0],
                    [1.0 / 12.0, 0.0, 5.0 / 400.0]]
    exp_output_2 = [[2.0 / 6.0, 1.0 / 5.0],
                    [0, 0],
                    [6.0 / 6.0, 5.0 / 5.0]]
    ioa_output_1, ioa_output_2 = self.execute(graph_fn, [])
    self.assertAllClose(ioa_output_1, exp_output_1)
    self.assertAllClose(ioa_output_2, exp_output_2)

  def test_prune_non_overlapping_boxes(self):
    def graph_fn():
      corners1 = tf.constant([[4.0, 3.0, 7.0, 5.0], [5.0, 6.0, 10.0, 7.0]])
      corners2 = tf.constant([[3.0, 4.0, 6.0, 8.0], [14.0, 14.0, 15.0, 15.0],
                              [0.0, 0.0, 20.0, 20.0]])
      boxes1 = box_list.BoxList(corners1)
      boxes2 = box_list.BoxList(corners2)
      minoverlap = 0.5

      exp_output_1 = boxes1
      exp_output_2 = box_list.BoxList(tf.constant(0.0, shape=[0, 4]))
      output_1, keep_indices_1 = box_list_ops.prune_non_overlapping_boxes(
          boxes1, boxes2, min_overlap=minoverlap)
      output_2, keep_indices_2 = box_list_ops.prune_non_overlapping_boxes(
          boxes2, boxes1, min_overlap=minoverlap)
      return (output_1.get(), keep_indices_1, output_2.get(), keep_indices_2,
              exp_output_1.get(), exp_output_2.get())

    (output_1_, keep_indices_1_, output_2_, keep_indices_2_, exp_output_1_,
     exp_output_2_) = self.execute_cpu(graph_fn, [])
    self.assertAllClose(output_1_, exp_output_1_)
    self.assertAllClose(output_2_, exp_output_2_)
    self.assertAllEqual(keep_indices_1_, [0, 1])
    self.assertAllEqual(keep_indices_2_, [])

  def test_prune_small_boxes(self):
    def graph_fn():
      boxes = tf.constant([[4.0, 3.0, 7.0, 5.0],
                           [5.0, 6.0, 10.0, 7.0],
                           [3.0, 4.0, 6.0, 8.0],
                           [14.0, 14.0, 15.0, 15.0],
                           [0.0, 0.0, 20.0, 20.0]])
      boxes = box_list.BoxList(boxes)
      pruned_boxes = box_list_ops.prune_small_boxes(boxes, 3)
      return pruned_boxes.get()
    exp_boxes = [[3.0, 4.0, 6.0, 8.0],
                 [0.0, 0.0, 20.0, 20.0]]
    pruned_boxes = self.execute(graph_fn, [])
    self.assertAllEqual(pruned_boxes, exp_boxes)

  def test_prune_small_boxes_prunes_boxes_with_negative_side(self):
    def graph_fn():
      boxes = tf.constant([[4.0, 3.0, 7.0, 5.0],
                           [5.0, 6.0, 10.0, 7.0],
                           [3.0, 4.0, 6.0, 8.0],
                           [14.0, 14.0, 15.0, 15.0],
                           [0.0, 0.0, 20.0, 20.0],
                           [2.0, 3.0, 1.5, 7.0],  # negative height
                           [2.0, 3.0, 5.0, 1.7]])  # negative width
      boxes = box_list.BoxList(boxes)
      pruned_boxes = box_list_ops.prune_small_boxes(boxes, 3)
      return pruned_boxes.get()
    exp_boxes = [[3.0, 4.0, 6.0, 8.0],
                 [0.0, 0.0, 20.0, 20.0]]
    pruned_boxes = self.execute_cpu(graph_fn, [])
    self.assertAllEqual(pruned_boxes, exp_boxes)

  def test_change_coordinate_frame(self):
    def graph_fn():
      corners = tf.constant([[0.25, 0.5, 0.75, 0.75], [0.5, 0.0, 1.0, 1.0]])
      window = tf.constant([0.25, 0.25, 0.75, 0.75])
      boxes = box_list.BoxList(corners)

      expected_corners = tf.constant([[0, 0.5, 1.0, 1.0],
                                      [0.5, -0.5, 1.5, 1.5]])
      expected_boxes = box_list.BoxList(expected_corners)
      output = box_list_ops.change_coordinate_frame(boxes, window)
      return output.get(), expected_boxes.get()
    output_, expected_boxes_ = self.execute(graph_fn, [])
    self.assertAllClose(output_, expected_boxes_)

  def test_ioaworks_on_empty_inputs(self):
    def graph_fn():
      corners1 = tf.constant([[4.0, 3.0, 7.0, 5.0], [5.0, 6.0, 10.0, 7.0]])
      corners2 = tf.constant([[3.0, 4.0, 6.0, 8.0], [14.0, 14.0, 15.0, 15.0],
                              [0.0, 0.0, 20.0, 20.0]])
      boxes1 = box_list.BoxList(corners1)
      boxes2 = box_list.BoxList(corners2)
      boxes_empty = box_list.BoxList(tf.zeros((0, 4)))
      ioa_empty_1 = box_list_ops.ioa(boxes1, boxes_empty)
      ioa_empty_2 = box_list_ops.ioa(boxes_empty, boxes2)
      ioa_empty_3 = box_list_ops.ioa(boxes_empty, boxes_empty)
      return ioa_empty_1, ioa_empty_2, ioa_empty_3
    ioa_output_1, ioa_output_2, ioa_output_3 = self.execute(graph_fn, [])
    self.assertAllEqual(ioa_output_1.shape, (2, 0))
    self.assertAllEqual(ioa_output_2.shape, (0, 3))
    self.assertAllEqual(ioa_output_3.shape, (0, 0))

  def test_pairwise_distances(self):
    def graph_fn():
      corners1 = tf.constant([[0.0, 0.0, 0.0, 0.0],
                              [1.0, 1.0, 0.0, 2.0]])
      corners2 = tf.constant([[3.0, 4.0, 1.0, 0.0],
                              [-4.0, 0.0, 0.0, 3.0],
                              [0.0, 0.0, 0.0, 0.0]])
      boxes1 = box_list.BoxList(corners1)
      boxes2 = box_list.BoxList(corners2)
      dist_matrix = box_list_ops.sq_dist(boxes1, boxes2)
      return dist_matrix
    exp_output = [[26, 25, 0], [18, 27, 6]]
    dist_output = self.execute(graph_fn, [])
    self.assertAllClose(dist_output, exp_output)

  def test_boolean_mask(self):
    def graph_fn():
      corners = tf.constant(
          [4 * [0.0], 4 * [1.0], 4 * [2.0], 4 * [3.0], 4 * [4.0]])
      indicator = tf.constant([True, False, True, False, True], tf.bool)
      boxes = box_list.BoxList(corners)
      subset = box_list_ops.boolean_mask(boxes, indicator)
      return subset.get()
    expected_subset = [4 * [0.0], 4 * [2.0], 4 * [4.0]]
    subset_output = self.execute_cpu(graph_fn, [])
    self.assertAllClose(subset_output, expected_subset)

  def test_static_boolean_mask_with_field(self):

    def graph_fn(corners, weights, indicator):
      boxes = box_list.BoxList(corners)
      boxes.add_field('weights', weights)
      subset = box_list_ops.boolean_mask(
          boxes,
          indicator, ['weights'],
          use_static_shapes=True,
          indicator_sum=3)
      return (subset.get_field('boxes'), subset.get_field('weights'))

    corners = np.array(
        [4 * [0.0], 4 * [1.0], 4 * [2.0], 4 * [3.0], 4 * [4.0]],
        dtype=np.float32)
    indicator = np.array([True, False, True, False, True], dtype=np.bool)
    weights = np.array([[.1], [.3], [.5], [.7], [.9]], dtype=np.float32)
    result_boxes, result_weights = self.execute_cpu(
        graph_fn, [corners, weights, indicator])
    expected_boxes = [4 * [0.0], 4 * [2.0], 4 * [4.0]]
    expected_weights = [[.1], [.5], [.9]]

    self.assertAllClose(result_boxes, expected_boxes)
    self.assertAllClose(result_weights, expected_weights)

  def test_gather(self):
    def graph_fn():
      corners = tf.constant(
          [4 * [0.0], 4 * [1.0], 4 * [2.0], 4 * [3.0], 4 * [4.0]])
      indices = tf.constant([0, 2, 4], tf.int32)
      boxes = box_list.BoxList(corners)
      subset = box_list_ops.gather(boxes, indices)
      return subset.get()
    expected_subset = [4 * [0.0], 4 * [2.0], 4 * [4.0]]
    subset_output = self.execute(graph_fn, [])
    self.assertAllClose(subset_output, expected_subset)

  def test_static_gather_with_field(self):

    def graph_fn(corners, weights, indices):
      boxes = box_list.BoxList(corners)
      boxes.add_field('weights', weights)
      subset = box_list_ops.gather(
          boxes, indices, ['weights'], use_static_shapes=True)
      return (subset.get_field('boxes'), subset.get_field('weights'))

    corners = np.array([4 * [0.0], 4 * [1.0], 4 * [2.0], 4 * [3.0],
                        4 * [4.0]], dtype=np.float32)
    weights = np.array([[.1], [.3], [.5], [.7], [.9]], dtype=np.float32)
    indices = np.array([0, 2, 4], dtype=np.int32)

    result_boxes, result_weights = self.execute(graph_fn,
                                                [corners, weights, indices])
    expected_boxes = [4 * [0.0], 4 * [2.0], 4 * [4.0]]
    expected_weights = [[.1], [.5], [.9]]
    self.assertAllClose(result_boxes, expected_boxes)
    self.assertAllClose(result_weights, expected_weights)

  def test_gather_with_invalid_field(self):
    corners = tf.constant([4 * [0.0], 4 * [1.0]])
    indices = tf.constant([0, 1], tf.int32)
    weights = tf.constant([[.1], [.3]], tf.float32)

    boxes = box_list.BoxList(corners)
    boxes.add_field('weights', weights)
    with self.assertRaises(ValueError):
      box_list_ops.gather(boxes, indices, ['foo', 'bar'])

  def test_gather_with_invalid_inputs(self):
    corners = tf.constant(
        [4 * [0.0], 4 * [1.0], 4 * [2.0], 4 * [3.0], 4 * [4.0]])
    indices_float32 = tf.constant([0, 2, 4], tf.float32)
    boxes = box_list.BoxList(corners)
    with self.assertRaises(ValueError):
      _ = box_list_ops.gather(boxes, indices_float32)
    indices_2d = tf.constant([[0, 2, 4]], tf.int32)
    boxes = box_list.BoxList(corners)
    with self.assertRaises(ValueError):
      _ = box_list_ops.gather(boxes, indices_2d)

  def test_gather_with_dynamic_indexing(self):
    def graph_fn():
      corners = tf.constant(
          [4 * [0.0], 4 * [1.0], 4 * [2.0], 4 * [3.0], 4 * [4.0]])
      weights = tf.constant([.5, .3, .7, .1, .9], tf.float32)
      indices = tf.reshape(tf.where(tf.greater(weights, 0.4)), [-1])
      boxes = box_list.BoxList(corners)
      boxes.add_field('weights', weights)
      subset = box_list_ops.gather(boxes, indices, ['weights'])
      return subset.get(), subset.get_field('weights')
    expected_subset = [4 * [0.0], 4 * [2.0], 4 * [4.0]]
    expected_weights = [.5, .7, .9]
    subset_output, weights_output = self.execute(graph_fn, [])
    self.assertAllClose(subset_output, expected_subset)
    self.assertAllClose(weights_output, expected_weights)

  def test_sort_by_field_ascending_order(self):
    exp_corners = [[0, 0, 1, 1], [0, 0.1, 1, 1.1], [0, -0.1, 1, 0.9],
                   [0, 10, 1, 11], [0, 10.1, 1, 11.1], [0, 100, 1, 101]]
    exp_scores = [.95, .9, .75, .6, .5, .3]
    exp_weights = [.2, .45, .6, .75, .8, .92]

    def graph_fn():
      shuffle = [2, 4, 0, 5, 1, 3]
      corners = tf.constant([exp_corners[i] for i in shuffle], tf.float32)
      boxes = box_list.BoxList(corners)
      boxes.add_field('scores', tf.constant(
          [exp_scores[i] for i in shuffle], tf.float32))
      boxes.add_field('weights', tf.constant(
          [exp_weights[i] for i in shuffle], tf.float32))
      sort_by_weight = box_list_ops.sort_by_field(
          boxes,
          'weights',
          order=box_list_ops.SortOrder.ascend)
      return [sort_by_weight.get(), sort_by_weight.get_field('scores'),
              sort_by_weight.get_field('weights')]
    corners_out, scores_out, weights_out = self.execute(graph_fn, [])
    self.assertAllClose(corners_out, exp_corners)
    self.assertAllClose(scores_out, exp_scores)
    self.assertAllClose(weights_out, exp_weights)

  def test_sort_by_field_descending_order(self):
    exp_corners = [[0, 0, 1, 1], [0, 0.1, 1, 1.1], [0, -0.1, 1, 0.9],
                   [0, 10, 1, 11], [0, 10.1, 1, 11.1], [0, 100, 1, 101]]
    exp_scores = [.95, .9, .75, .6, .5, .3]
    exp_weights = [.2, .45, .6, .75, .8, .92]

    def graph_fn():
      shuffle = [2, 4, 0, 5, 1, 3]
      corners = tf.constant([exp_corners[i] for i in shuffle], tf.float32)
      boxes = box_list.BoxList(corners)
      boxes.add_field('scores', tf.constant(
          [exp_scores[i] for i in shuffle], tf.float32))
      boxes.add_field('weights', tf.constant(
          [exp_weights[i] for i in shuffle], tf.float32))
      sort_by_score = box_list_ops.sort_by_field(boxes, 'scores')
      return (sort_by_score.get(), sort_by_score.get_field('scores'),
              sort_by_score.get_field('weights'))

    corners_out, scores_out, weights_out = self.execute(graph_fn, [])
    self.assertAllClose(corners_out, exp_corners)
    self.assertAllClose(scores_out, exp_scores)
    self.assertAllClose(weights_out, exp_weights)

  def test_sort_by_field_invalid_inputs(self):
    corners = tf.constant([4 * [0.0], 4 * [0.5], 4 * [1.0], 4 * [2.0], 4 *
                           [3.0], 4 * [4.0]])
    misc = tf.constant([[.95, .9], [.5, .3]], tf.float32)
    weights = tf.constant([[.1, .2]], tf.float32)
    boxes = box_list.BoxList(corners)
    boxes.add_field('misc', misc)
    boxes.add_field('weights', weights)

    with self.assertRaises(ValueError):
      box_list_ops.sort_by_field(boxes, 'area')

    with self.assertRaises(ValueError):
      box_list_ops.sort_by_field(boxes, 'misc')

    with self.assertRaises(ValueError):
      box_list_ops.sort_by_field(boxes, 'weights')

  def test_visualize_boxes_in_image(self):
    def graph_fn():
      image = tf.zeros((6, 4, 3))
      corners = tf.constant([[0, 0, 5, 3],
                             [0, 0, 3, 2]], tf.float32)
      boxes = box_list.BoxList(corners)
      image_and_boxes = box_list_ops.visualize_boxes_in_image(image, boxes)
      image_and_boxes_bw = tf.cast(
          tf.greater(tf.reduce_sum(image_and_boxes, 2), 0.0), dtype=tf.float32)
      return image_and_boxes_bw
    exp_result = [[1, 1, 1, 0],
                  [1, 1, 1, 0],
                  [1, 1, 1, 0],
                  [1, 0, 1, 0],
                  [1, 1, 1, 0],
                  [0, 0, 0, 0]]
    output = self.execute_cpu(graph_fn, [])
    self.assertAllEqual(output.astype(int), exp_result)

  def test_filter_field_value_equals(self):
    def graph_fn():
      corners = tf.constant([[0, 0, 1, 1],
                             [0, 0.1, 1, 1.1],
                             [0, -0.1, 1, 0.9],
                             [0, 10, 1, 11],
                             [0, 10.1, 1, 11.1],
                             [0, 100, 1, 101]], tf.float32)
      boxes = box_list.BoxList(corners)
      boxes.add_field('classes', tf.constant([1, 2, 1, 2, 2, 1]))
      filtered_boxes1 = box_list_ops.filter_field_value_equals(
          boxes, 'classes', 1)
      filtered_boxes2 = box_list_ops.filter_field_value_equals(
          boxes, 'classes', 2)
      return filtered_boxes1.get(), filtered_boxes2.get()
    exp_output1 = [[0, 0, 1, 1], [0, -0.1, 1, 0.9], [0, 100, 1, 101]]
    exp_output2 = [[0, 0.1, 1, 1.1], [0, 10, 1, 11], [0, 10.1, 1, 11.1]]
    filtered_output1, filtered_output2 = self.execute_cpu(graph_fn, [])
    self.assertAllClose(filtered_output1, exp_output1)
    self.assertAllClose(filtered_output2, exp_output2)

  def test_filter_greater_than(self):
    def graph_fn():
      corners = tf.constant([[0, 0, 1, 1],
                             [0, 0.1, 1, 1.1],
                             [0, -0.1, 1, 0.9],
                             [0, 10, 1, 11],
                             [0, 10.1, 1, 11.1],
                             [0, 100, 1, 101]], tf.float32)
      boxes = box_list.BoxList(corners)
      boxes.add_field('scores', tf.constant([.1, .75, .9, .5, .5, .8]))
      thresh = .6
      filtered_boxes = box_list_ops.filter_greater_than(boxes, thresh)
      return filtered_boxes.get()
    exp_output = [[0, 0.1, 1, 1.1], [0, -0.1, 1, 0.9], [0, 100, 1, 101]]
    filtered_output = self.execute_cpu(graph_fn, [])
    self.assertAllClose(filtered_output, exp_output)

  def test_clip_box_list(self):
    def graph_fn():
      boxlist = box_list.BoxList(
          tf.constant([[0.1, 0.1, 0.4, 0.4], [0.1, 0.1, 0.5, 0.5],
                       [0.6, 0.6, 0.8, 0.8], [0.2, 0.2, 0.3, 0.3]], tf.float32))
      boxlist.add_field('classes', tf.constant([0, 0, 1, 1]))
      boxlist.add_field('scores', tf.constant([0.75, 0.65, 0.3, 0.2]))
      num_boxes = 2
      clipped_boxlist = box_list_ops.pad_or_clip_box_list(boxlist, num_boxes)
      return (clipped_boxlist.get(), clipped_boxlist.get_field('classes'),
              clipped_boxlist.get_field('scores'))

    expected_boxes = [[0.1, 0.1, 0.4, 0.4], [0.1, 0.1, 0.5, 0.5]]
    expected_classes = [0, 0]
    expected_scores = [0.75, 0.65]
    boxes_out, classes_out, scores_out = self.execute(graph_fn, [])

    self.assertAllClose(expected_boxes, boxes_out)
    self.assertAllEqual(expected_classes, classes_out)
    self.assertAllClose(expected_scores, scores_out)

  def test_pad_box_list(self):
    def graph_fn():
      boxlist = box_list.BoxList(
          tf.constant([[0.1, 0.1, 0.4, 0.4], [0.1, 0.1, 0.5, 0.5]], tf.float32))
      boxlist.add_field('classes', tf.constant([0, 1]))
      boxlist.add_field('scores', tf.constant([0.75, 0.2]))
      num_boxes = 4
      padded_boxlist = box_list_ops.pad_or_clip_box_list(boxlist, num_boxes)
      return (padded_boxlist.get(), padded_boxlist.get_field('classes'),
              padded_boxlist.get_field('scores'))
    expected_boxes = [[0.1, 0.1, 0.4, 0.4], [0.1, 0.1, 0.5, 0.5],
                      [0, 0, 0, 0], [0, 0, 0, 0]]
    expected_classes = [0, 1, 0, 0]
    expected_scores = [0.75, 0.2, 0, 0]
    boxes_out, classes_out, scores_out = self.execute(graph_fn, [])
    self.assertAllClose(expected_boxes, boxes_out)
    self.assertAllEqual(expected_classes, classes_out)
    self.assertAllClose(expected_scores, scores_out)

  def test_select_random_box(self):
    boxes = [[0., 0., 1., 1.],
             [0., 1., 2., 3.],
             [0., 2., 3., 4.]]
    def graph_fn():
      corners = tf.constant(boxes, dtype=tf.float32)
      boxlist = box_list.BoxList(corners)
      random_bbox, valid = box_list_ops.select_random_box(boxlist)
      return random_bbox, valid
    random_bbox_out, valid_out = self.execute(graph_fn, [])
    norm_small = any(
        [np.linalg.norm(random_bbox_out - box) < 1e-6 for box in boxes])
    self.assertTrue(norm_small)
    self.assertTrue(valid_out)

  def test_select_random_box_with_empty_boxlist(self):
    def graph_fn():
      corners = tf.constant([], shape=[0, 4], dtype=tf.float32)
      boxlist = box_list.BoxList(corners)
      random_bbox, valid = box_list_ops.select_random_box(boxlist)
      return random_bbox, valid
    random_bbox_out, valid_out = self.execute_cpu(graph_fn, [])
    expected_bbox_out = np.array([[-1., -1., -1., -1.]], dtype=np.float32)
    self.assertAllEqual(expected_bbox_out, random_bbox_out)
    self.assertFalse(valid_out)

  def test_get_minimal_coverage_box(self):
    def graph_fn():
      boxes = [[0., 0., 1., 1.],
               [-1., 1., 2., 3.],
               [0., 2., 3., 4.]]
      corners = tf.constant(boxes, dtype=tf.float32)
      boxlist = box_list.BoxList(corners)
      coverage_box = box_list_ops.get_minimal_coverage_box(boxlist)
      return coverage_box
    coverage_box_out = self.execute(graph_fn, [])
    expected_coverage_box = [[-1., 0., 3., 4.]]
    self.assertAllClose(expected_coverage_box, coverage_box_out)

  def test_get_minimal_coverage_box_with_empty_boxlist(self):
    def graph_fn():
      corners = tf.constant([], shape=[0, 4], dtype=tf.float32)
      boxlist = box_list.BoxList(corners)
      coverage_box = box_list_ops.get_minimal_coverage_box(boxlist)
      return coverage_box
    coverage_box_out = self.execute(graph_fn, [])
    self.assertAllClose([[0.0, 0.0, 1.0, 1.0]], coverage_box_out)


class ConcatenateTest(test_case.TestCase):

  def test_invalid_input_box_list_list(self):
    with self.assertRaises(ValueError):
      box_list_ops.concatenate(None)
    with self.assertRaises(ValueError):
      box_list_ops.concatenate([])
    with self.assertRaises(ValueError):
      corners = tf.constant([[0, 0, 0, 0]], tf.float32)
      boxlist = box_list.BoxList(corners)
      box_list_ops.concatenate([boxlist, 2])

  def test_concatenate_with_missing_fields(self):
    corners1 = tf.constant([[0, 0, 0, 0], [1, 2, 3, 4]], tf.float32)
    scores1 = tf.constant([1.0, 2.1])
    corners2 = tf.constant([[0, 3, 1, 6], [2, 4, 3, 8]], tf.float32)
    boxlist1 = box_list.BoxList(corners1)
    boxlist1.add_field('scores', scores1)
    boxlist2 = box_list.BoxList(corners2)
    with self.assertRaises(ValueError):
      box_list_ops.concatenate([boxlist1, boxlist2])

  def test_concatenate_with_incompatible_field_shapes(self):
    corners1 = tf.constant([[0, 0, 0, 0], [1, 2, 3, 4]], tf.float32)
    scores1 = tf.constant([1.0, 2.1])
    corners2 = tf.constant([[0, 3, 1, 6], [2, 4, 3, 8]], tf.float32)
    scores2 = tf.constant([[1.0, 1.0], [2.1, 3.2]])
    boxlist1 = box_list.BoxList(corners1)
    boxlist1.add_field('scores', scores1)
    boxlist2 = box_list.BoxList(corners2)
    boxlist2.add_field('scores', scores2)
    with self.assertRaises(ValueError):
      box_list_ops.concatenate([boxlist1, boxlist2])

  def test_concatenate_is_correct(self):
    def graph_fn():
      corners1 = tf.constant([[0, 0, 0, 0], [1, 2, 3, 4]], tf.float32)
      scores1 = tf.constant([1.0, 2.1])
      corners2 = tf.constant([[0, 3, 1, 6], [2, 4, 3, 8], [1, 0, 5, 10]],
                             tf.float32)
      scores2 = tf.constant([1.0, 2.1, 5.6])
      boxlist1 = box_list.BoxList(corners1)
      boxlist1.add_field('scores', scores1)
      boxlist2 = box_list.BoxList(corners2)
      boxlist2.add_field('scores', scores2)
      result = box_list_ops.concatenate([boxlist1, boxlist2])
      return result.get(), result.get_field('scores')
    exp_corners = [[0, 0, 0, 0],
                   [1, 2, 3, 4],
                   [0, 3, 1, 6],
                   [2, 4, 3, 8],
                   [1, 0, 5, 10]]
    exp_scores = [1.0, 2.1, 1.0, 2.1, 5.6]
    corners_output, scores_output = self.execute(graph_fn, [])
    self.assertAllClose(corners_output, exp_corners)
    self.assertAllClose(scores_output, exp_scores)


class NonMaxSuppressionTest(test_case.TestCase):

  def test_select_from_three_clusters(self):
    def graph_fn():
      corners = tf.constant([[0, 0, 1, 1],
                             [0, 0.1, 1, 1.1],
                             [0, -0.1, 1, 0.9],
                             [0, 10, 1, 11],
                             [0, 10.1, 1, 11.1],
                             [0, 100, 1, 101]], tf.float32)
      boxes = box_list.BoxList(corners)
      boxes.add_field('scores', tf.constant([.9, .75, .6, .95, .5, .3]))
      iou_thresh = .5
      max_output_size = 3
      nms = box_list_ops.non_max_suppression(
          boxes, iou_thresh, max_output_size)
      return nms.get()
    exp_nms = [[0, 10, 1, 11],
               [0, 0, 1, 1],
               [0, 100, 1, 101]]
    nms_output = self.execute_cpu(graph_fn, [])
    self.assertAllClose(nms_output, exp_nms)

  def test_select_at_most_two_boxes_from_three_clusters(self):
    def graph_fn():
      corners = tf.constant([[0, 0, 1, 1],
                             [0, 0.1, 1, 1.1],
                             [0, -0.1, 1, 0.9],
                             [0, 10, 1, 11],
                             [0, 10.1, 1, 11.1],
                             [0, 100, 1, 101]], tf.float32)
      boxes = box_list.BoxList(corners)
      boxes.add_field('scores', tf.constant([.9, .75, .6, .95, .5, .3]))
      iou_thresh = .5
      max_output_size = 2
      nms = box_list_ops.non_max_suppression(
          boxes, iou_thresh, max_output_size)
      return nms.get()
    exp_nms = [[0, 10, 1, 11],
               [0, 0, 1, 1]]
    nms_output = self.execute_cpu(graph_fn, [])
    self.assertAllClose(nms_output, exp_nms)

  def test_select_at_most_thirty_boxes_from_three_clusters(self):
    def graph_fn():
      corners = tf.constant([[0, 0, 1, 1],
                             [0, 0.1, 1, 1.1],
                             [0, -0.1, 1, 0.9],
                             [0, 10, 1, 11],
                             [0, 10.1, 1, 11.1],
                             [0, 100, 1, 101]], tf.float32)
      boxes = box_list.BoxList(corners)
      boxes.add_field('scores', tf.constant([.9, .75, .6, .95, .5, .3]))
      iou_thresh = .5
      max_output_size = 30
      nms = box_list_ops.non_max_suppression(
          boxes, iou_thresh, max_output_size)
      return nms.get()
    exp_nms = [[0, 10, 1, 11],
               [0, 0, 1, 1],
               [0, 100, 1, 101]]
    nms_output = self.execute_cpu(graph_fn, [])
    self.assertAllClose(nms_output, exp_nms)

  def test_select_single_box(self):
    def graph_fn():
      corners = tf.constant([[0, 0, 1, 1]], tf.float32)
      boxes = box_list.BoxList(corners)
      boxes.add_field('scores', tf.constant([.9]))
      iou_thresh = .5
      max_output_size = 3
      nms = box_list_ops.non_max_suppression(
          boxes, iou_thresh, max_output_size)
      return nms.get()
    exp_nms = [[0, 0, 1, 1]]
    nms_output = self.execute_cpu(graph_fn, [])
    self.assertAllClose(nms_output, exp_nms)

  def test_select_from_ten_identical_boxes(self):
    def graph_fn():
      corners = tf.constant(10 * [[0, 0, 1, 1]], tf.float32)
      boxes = box_list.BoxList(corners)
      boxes.add_field('scores', tf.constant(10 * [.9]))
      iou_thresh = .5
      max_output_size = 3
      nms = box_list_ops.non_max_suppression(
          boxes, iou_thresh, max_output_size)
      return nms.get()
    exp_nms = [[0, 0, 1, 1]]
    nms_output = self.execute_cpu(graph_fn, [])
    self.assertAllClose(nms_output, exp_nms)

  def test_copy_extra_fields(self):
    tensor1 = np.array([[1], [4]])
    tensor2 = np.array([[1, 1], [2, 2]])
    def graph_fn():
      corners = tf.constant([[0, 0, 1, 1],
                             [0, 0.1, 1, 1.1]], tf.float32)
      boxes = box_list.BoxList(corners)

      boxes.add_field('tensor1', tf.constant(tensor1))
      boxes.add_field('tensor2', tf.constant(tensor2))
      new_boxes = box_list.BoxList(tf.constant([[0, 0, 10, 10],
                                                [1, 3, 5, 5]], tf.float32))
      new_boxes = box_list_ops._copy_extra_fields(new_boxes, boxes)
      return new_boxes.get_field('tensor1'), new_boxes.get_field('tensor2')
    tensor1_out, tensor2_out = self.execute_cpu(graph_fn, [])
    self.assertAllClose(tensor1, tensor1_out)
    self.assertAllClose(tensor2, tensor2_out)


class CoordinatesConversionTest(test_case.TestCase):

  def test_to_normalized_coordinates(self):
    def graph_fn():
      coordinates = tf.constant([[0, 0, 100, 100],
                                 [25, 25, 75, 75]], tf.float32)
      img = tf.ones((128, 100, 100, 3))
      boxlist = box_list.BoxList(coordinates)
      normalized_boxlist = box_list_ops.to_normalized_coordinates(
          boxlist, tf.shape(img)[1], tf.shape(img)[2])
      return normalized_boxlist.get()
    expected_boxes = [[0, 0, 1, 1],
                      [0.25, 0.25, 0.75, 0.75]]
    normalized_boxes = self.execute(graph_fn, [])
    self.assertAllClose(normalized_boxes, expected_boxes)

  def test_to_normalized_coordinates_already_normalized(self):
    def graph_fn():
      coordinates = tf.constant([[0, 0, 1, 1],
                                 [0.25, 0.25, 0.75, 0.75]], tf.float32)
      img = tf.ones((128, 100, 100, 3))
      boxlist = box_list.BoxList(coordinates)
      normalized_boxlist = box_list_ops.to_normalized_coordinates(
          boxlist, tf.shape(img)[1], tf.shape(img)[2])
      return normalized_boxlist.get()
    with self.assertRaisesOpError('assertion failed'):
      self.execute_cpu(graph_fn, [])

  def test_to_absolute_coordinates(self):
    def graph_fn():
      coordinates = tf.constant([[0, 0, 1, 1],
                                 [0.25, 0.25, 0.75, 0.75]], tf.float32)
      img = tf.ones((128, 100, 100, 3))
      boxlist = box_list.BoxList(coordinates)
      absolute_boxlist = box_list_ops.to_absolute_coordinates(boxlist,
                                                              tf.shape(img)[1],
                                                              tf.shape(img)[2])
      return absolute_boxlist.get()
    expected_boxes = [[0, 0, 100, 100],
                      [25, 25, 75, 75]]
    absolute_boxes = self.execute(graph_fn, [])
    self.assertAllClose(absolute_boxes, expected_boxes)

  def test_to_absolute_coordinates_already_abolute(self):
    def graph_fn():
      coordinates = tf.constant([[0, 0, 100, 100],
                                 [25, 25, 75, 75]], tf.float32)
      img = tf.ones((128, 100, 100, 3))
      boxlist = box_list.BoxList(coordinates)
      absolute_boxlist = box_list_ops.to_absolute_coordinates(boxlist,
                                                              tf.shape(img)[1],
                                                              tf.shape(img)[2])
      return absolute_boxlist.get()
    with self.assertRaisesOpError('assertion failed'):
      self.execute_cpu(graph_fn, [])

  def test_convert_to_normalized_and_back(self):
    coordinates = np.random.uniform(size=(100, 4))
    coordinates = np.round(np.sort(coordinates) * 200)
    coordinates[:, 2:4] += 1
    coordinates[99, :] = [0, 0, 201, 201]
    def graph_fn():
      img = tf.ones((128, 202, 202, 3))

      boxlist = box_list.BoxList(tf.constant(coordinates, tf.float32))
      boxlist = box_list_ops.to_normalized_coordinates(boxlist,
                                                       tf.shape(img)[1],
                                                       tf.shape(img)[2])
      boxlist = box_list_ops.to_absolute_coordinates(boxlist,
                                                     tf.shape(img)[1],
                                                     tf.shape(img)[2])
      return boxlist.get()
    out = self.execute(graph_fn, [])
    self.assertAllClose(out, coordinates)

  def test_convert_to_absolute_and_back(self):
    coordinates = np.random.uniform(size=(100, 4))
    coordinates = np.sort(coordinates)
    coordinates[99, :] = [0, 0, 1, 1]
    def graph_fn():
      img = tf.ones((128, 202, 202, 3))
      boxlist = box_list.BoxList(tf.constant(coordinates, tf.float32))
      boxlist = box_list_ops.to_absolute_coordinates(boxlist,
                                                     tf.shape(img)[1],
                                                     tf.shape(img)[2])
      boxlist = box_list_ops.to_normalized_coordinates(boxlist,
                                                       tf.shape(img)[1],
                                                       tf.shape(img)[2])
      return boxlist.get()
    out = self.execute(graph_fn, [])
    self.assertAllClose(out, coordinates)

  def test_to_absolute_coordinates_maximum_coordinate_check(self):
    def graph_fn():
      coordinates = tf.constant([[0, 0, 1.2, 1.2],
                                 [0.25, 0.25, 0.75, 0.75]], tf.float32)
      img = tf.ones((128, 100, 100, 3))
      boxlist = box_list.BoxList(coordinates)
      absolute_boxlist = box_list_ops.to_absolute_coordinates(
          boxlist,
          tf.shape(img)[1],
          tf.shape(img)[2],
          maximum_normalized_coordinate=1.1)
      return absolute_boxlist.get()
    with self.assertRaisesOpError('assertion failed'):
      self.execute_cpu(graph_fn, [])


class BoxRefinementTest(test_case.TestCase):

  def test_box_voting(self):
    def graph_fn():
      candidates = box_list.BoxList(
          tf.constant([[0.1, 0.1, 0.4, 0.4], [0.6, 0.6, 0.8, 0.8]], tf.float32))
      candidates.add_field('ExtraField', tf.constant([1, 2]))
      pool = box_list.BoxList(
          tf.constant([[0.1, 0.1, 0.4, 0.4], [0.1, 0.1, 0.5, 0.5],
                       [0.6, 0.6, 0.8, 0.8]], tf.float32))
      pool.add_field('scores', tf.constant([0.75, 0.25, 0.3]))
      averaged_boxes = box_list_ops.box_voting(candidates, pool)
      return (averaged_boxes.get(), averaged_boxes.get_field('scores'),
              averaged_boxes.get_field('ExtraField'))

    expected_boxes = [[0.1, 0.1, 0.425, 0.425], [0.6, 0.6, 0.8, 0.8]]
    expected_scores = [0.5, 0.3]
    boxes_out, scores_out, extra_field_out = self.execute(graph_fn, [])
    self.assertAllClose(expected_boxes, boxes_out)
    self.assertAllClose(expected_scores, scores_out)
    self.assertAllEqual(extra_field_out, [1, 2])

  def test_box_voting_fails_with_negative_scores(self):
    def graph_fn():
      candidates = box_list.BoxList(
          tf.constant([[0.1, 0.1, 0.4, 0.4]], tf.float32))
      pool = box_list.BoxList(tf.constant([[0.1, 0.1, 0.4, 0.4]], tf.float32))
      pool.add_field('scores', tf.constant([-0.2]))
      averaged_boxes = box_list_ops.box_voting(candidates, pool)
      return averaged_boxes.get()

    with self.assertRaisesOpError('Scores must be non negative'):
      self.execute_cpu(graph_fn, [])

  def test_box_voting_fails_when_unmatched(self):
    def graph_fn():
      candidates = box_list.BoxList(
          tf.constant([[0.1, 0.1, 0.4, 0.4]], tf.float32))
      pool = box_list.BoxList(tf.constant([[0.6, 0.6, 0.8, 0.8]], tf.float32))
      pool.add_field('scores', tf.constant([0.2]))
      averaged_boxes = box_list_ops.box_voting(candidates, pool)
      return averaged_boxes.get()
    with self.assertRaisesOpError('Each box in selected_boxes must match '
                                  'with at least one box in pool_boxes.'):
      self.execute_cpu(graph_fn, [])

  def test_refine_boxes(self):
    def graph_fn():
      pool = box_list.BoxList(
          tf.constant([[0.1, 0.1, 0.4, 0.4], [0.1, 0.1, 0.5, 0.5],
                       [0.6, 0.6, 0.8, 0.8]], tf.float32))
      pool.add_field('ExtraField', tf.constant([1, 2, 3]))
      pool.add_field('scores', tf.constant([0.75, 0.25, 0.3]))
      averaged_boxes = box_list_ops.refine_boxes(pool, 0.5, 10)
      return (averaged_boxes.get(), averaged_boxes.get_field('scores'),
              averaged_boxes.get_field('ExtraField'))
    boxes_out, scores_out, extra_field_out = self.execute_cpu(graph_fn, [])
    expected_boxes = [[0.1, 0.1, 0.425, 0.425], [0.6, 0.6, 0.8, 0.8]]
    expected_scores = [0.5, 0.3]
    self.assertAllClose(expected_boxes, boxes_out)
    self.assertAllClose(expected_scores, scores_out)
    self.assertAllEqual(extra_field_out, [1, 3])

  def test_refine_boxes_multi_class(self):
    def graph_fn():
      pool = box_list.BoxList(
          tf.constant([[0.1, 0.1, 0.4, 0.4], [0.1, 0.1, 0.5, 0.5],
                       [0.6, 0.6, 0.8, 0.8], [0.2, 0.2, 0.3, 0.3]], tf.float32))
      pool.add_field('classes', tf.constant([0, 0, 1, 1]))
      pool.add_field('scores', tf.constant([0.75, 0.25, 0.3, 0.2]))
      averaged_boxes = box_list_ops.refine_boxes_multi_class(pool, 3, 0.5, 10)
      return (averaged_boxes.get(), averaged_boxes.get_field('scores'),
              averaged_boxes.get_field('classes'))
    boxes_out, scores_out, extra_field_out = self.execute_cpu(graph_fn, [])
    expected_boxes = [[0.1, 0.1, 0.425, 0.425], [0.6, 0.6, 0.8, 0.8],
                      [0.2, 0.2, 0.3, 0.3]]
    expected_scores = [0.5, 0.3, 0.2]
    self.assertAllClose(expected_boxes, boxes_out)
    self.assertAllClose(expected_scores, scores_out)
    self.assertAllEqual(extra_field_out, [0, 1, 1])

  def test_sample_boxes_by_jittering(self):
    def graph_fn():
      boxes = box_list.BoxList(
          tf.constant([[0.1, 0.1, 0.4, 0.4],
                       [0.1, 0.1, 0.5, 0.5],
                       [0.6, 0.6, 0.8, 0.8],
                       [0.2, 0.2, 0.3, 0.3]], tf.float32))
      sampled_boxes = box_list_ops.sample_boxes_by_jittering(
          boxlist=boxes, num_boxes_to_sample=10)
      iou = box_list_ops.iou(boxes, sampled_boxes)
      iou_max = tf.reduce_max(iou, axis=0)
      return sampled_boxes.get(), iou_max
    np_sampled_boxes, np_iou_max = self.execute(graph_fn, [])
    self.assertAllEqual(np_sampled_boxes.shape, [10, 4])
    self.assertAllGreater(np_iou_max, 0.3)


if __name__ == '__main__':
  tf.test.main()



# Copyright (c) Facebook, Inc. and its affiliates.
from typing import List
import fvcore.nn.weight_init as weight_init
import torch
from torch import nn
from torch.nn import functional as F

from detectron2.config import configurable
from detectron2.layers import Conv2d, ConvTranspose2d, ShapeSpec, cat, get_norm
from detectron2.structures import Instances
from detectron2.utils.events import get_event_storage
from detectron2.utils.registry import Registry

__all__ = [
    "BaseMaskRCNNHead",
    "MaskRCNNConvUpsampleHead",
    "build_mask_head",
    "ROI_MASK_HEAD_REGISTRY",
]


ROI_MASK_HEAD_REGISTRY = Registry("ROI_MASK_HEAD")
ROI_MASK_HEAD_REGISTRY.__doc__ = """
Registry for mask heads, which predicts instance masks given
per-region features.

The registered object will be called with `obj(cfg, input_shape)`.
"""


@torch.jit.unused
def mask_rcnn_loss(pred_mask_logits: torch.Tensor, instances: List[Instances], vis_period: int = 0):
    """
    Compute the mask prediction loss defined in the Mask R-CNN paper.

    Args:
        pred_mask_logits (Tensor): A tensor of shape (B, C, Hmask, Wmask) or (B, 1, Hmask, Wmask)
            for class-specific or class-agnostic, where B is the total number of predicted masks
            in all images, C is the number of foreground classes, and Hmask, Wmask are the height
            and width of the mask predictions. The values are logits.
        instances (list[Instances]): A list of N Instances, where N is the number of images
            in the batch. These instances are in 1:1
            correspondence with the pred_mask_logits. The ground-truth labels (class, box, mask,
            ...) associated with each instance are stored in fields.
        vis_period (int): the period (in steps) to dump visualization.

    Returns:
        mask_loss (Tensor): A scalar tensor containing the loss.
    """
    cls_agnostic_mask = pred_mask_logits.size(1) == 1
    total_num_masks = pred_mask_logits.size(0)
    mask_side_len = pred_mask_logits.size(2)
    assert pred_mask_logits.size(2) == pred_mask_logits.size(3), "Mask prediction must be square!"

    gt_classes = []
    gt_masks = []
    for instances_per_image in instances:
        if len(instances_per_image) == 0:
            continue
        if not cls_agnostic_mask:
            gt_classes_per_image = instances_per_image.gt_classes.to(dtype=torch.int64)
            gt_classes.append(gt_classes_per_image)

        gt_masks_per_image = instances_per_image.gt_masks.crop_and_resize(
            instances_per_image.proposal_boxes.tensor, mask_side_len
        ).to(device=pred_mask_logits.device)
        # A tensor of shape (N, M, M), N=#instances in the image; M=mask_side_len
        gt_masks.append(gt_masks_per_image)

    if len(gt_masks) == 0:
        return pred_mask_logits.sum() * 0

    gt_masks = cat(gt_masks, dim=0)

    if cls_agnostic_mask:
        pred_mask_logits = pred_mask_logits[:, 0]
    else:
        indices = torch.arange(total_num_masks)
        gt_classes = cat(gt_classes, dim=0)
        pred_mask_logits = pred_mask_logits[indices, gt_classes]

    if gt_masks.dtype == torch.bool:
        gt_masks_bool = gt_masks
    else:
        # Here we allow gt_masks to be float as well (depend on the implementation of rasterize())
        gt_masks_bool = gt_masks > 0.5
    gt_masks = gt_masks.to(dtype=torch.float32)

    # Log the training accuracy (using gt classes and 0.5 threshold)
    mask_incorrect = (pred_mask_logits > 0.0) != gt_masks_bool
    mask_accuracy = 1 - (mask_incorrect.sum().item() / max(mask_incorrect.numel(), 1.0))
    num_positive = gt_masks_bool.sum().item()
    false_positive = (mask_incorrect & ~gt_masks_bool).sum().item() / max(
        gt_masks_bool.numel() - num_positive, 1.0
    )
    false_negative = (mask_incorrect & gt_masks_bool).sum().item() / max(num_positive, 1.0)

    storage = get_event_storage()
    storage.put_scalar("mask_rcnn/accuracy", mask_accuracy)
    storage.put_scalar("mask_rcnn/false_positive", false_positive)
    storage.put_scalar("mask_rcnn/false_negative", false_negative)
    if vis_period > 0 and storage.iter % vis_period == 0:
        pred_masks = pred_mask_logits.sigmoid()
        vis_masks = torch.cat([pred_masks, gt_masks], axis=2)
        name = "Left: mask prediction;   Right: mask GT"
        for idx, vis_mask in enumerate(vis_masks):
            vis_mask = torch.stack([vis_mask] * 3, axis=0)
            storage.put_image(name + f" ({idx})", vis_mask)

    mask_loss = F.binary_cross_entropy_with_logits(pred_mask_logits, gt_masks, reduction="mean")
    return mask_loss


def mask_rcnn_inference(pred_mask_logits: torch.Tensor, pred_instances: List[Instances]):
    """
    Convert pred_mask_logits to estimated foreground probability masks while also
    extracting only the masks for the predicted classes in pred_instances. For each
    predicted box, the mask of the same class is attached to the instance by adding a
    new "pred_masks" field to pred_instances.

    Args:
        pred_mask_logits (Tensor): A tensor of shape (B, C, Hmask, Wmask) or (B, 1, Hmask, Wmask)
            for class-specific or class-agnostic, where B is the total number of predicted masks
            in all images, C is the number of foreground classes, and Hmask, Wmask are the height
            and width of the mask predictions. The values are logits.
        pred_instances (list[Instances]): A list of N Instances, where N is the number of images
            in the batch. Each Instances must have field "pred_classes".

    Returns:
        None. pred_instances will contain an extra "pred_masks" field storing a mask of size (Hmask,
            Wmask) for predicted class. Note that the masks are returned as a soft (non-quantized)
            masks the resolution predicted by the network; post-processing steps, such as resizing
            the predicted masks to the original image resolution and/or binarizing them, is left
            to the caller.
    """
    cls_agnostic_mask = pred_mask_logits.size(1) == 1

    if cls_agnostic_mask:
        mask_probs_pred = pred_mask_logits.sigmoid()
    else:
        # Select masks corresponding to the predicted classes
        num_masks = pred_mask_logits.shape[0]
        class_pred = cat([i.pred_classes for i in pred_instances])
        indices = torch.arange(num_masks, device=class_pred.device)
        mask_probs_pred = pred_mask_logits[indices, class_pred][:, None].sigmoid()
    # mask_probs_pred.shape: (B, 1, Hmask, Wmask)

    num_boxes_per_image = [len(i) for i in pred_instances]
    mask_probs_pred = mask_probs_pred.split(num_boxes_per_image, dim=0)

    for prob, instances in zip(mask_probs_pred, pred_instances):
        instances.pred_masks = prob  # (1, Hmask, Wmask)


class BaseMaskRCNNHead(nn.Module):
    """
    Implement the basic Mask R-CNN losses and inference logic described in :paper:`Mask R-CNN`
    """

    @configurable
    def __init__(self, *, loss_weight: float = 1.0, vis_period: int = 0):
        """
        NOTE: this interface is experimental.

        Args:
            loss_weight (float): multiplier of the loss
            vis_period (int): visualization period
        """
        super().__init__()
        self.vis_period = vis_period
        self.loss_weight = loss_weight

    @classmethod
    def from_config(cls, cfg, input_shape):
        return {"vis_period": cfg.VIS_PERIOD}

    def forward(self, x, instances: List[Instances]):
        """
        Args:
            x: input region feature(s) provided by :class:`ROIHeads`.
            instances (list[Instances]): contains the boxes & labels corresponding
                to the input features.
                Exact format is up to its caller to decide.
                Typically, this is the foreground instances in training, with
                "proposal_boxes" field and other gt annotations.
                In inference, it contains boxes that are already predicted.

        Returns:
            A dict of losses in training. The predicted "instances" in inference.
        """
        x = self.layers(x)
        if self.training:
            return {"loss_mask": mask_rcnn_loss(x, instances, self.vis_period) * self.loss_weight}
        else:
            mask_rcnn_inference(x, instances)
            return instances

    def layers(self, x):
        """
        Neural network layers that makes predictions from input features.
        """
        raise NotImplementedError


# To get torchscript support, we make the head a subclass of `nn.Sequential`.
# Therefore, to add new layers in this head class, please make sure they are
# added in the order they will be used in forward().
@ROI_MASK_HEAD_REGISTRY.register()
class MaskRCNNConvUpsampleHead(BaseMaskRCNNHead, nn.Sequential):
    """
    A mask head with several conv layers, plus an upsample layer (with `ConvTranspose2d`).
    Predictions are made with a final 1x1 conv layer.
    """

    @configurable
    def __init__(self, input_shape: ShapeSpec, *, num_classes, conv_dims, conv_norm="", **kwargs):
        """
        NOTE: this interface is experimental.

        Args:
            input_shape (ShapeSpec): shape of the input feature
            num_classes (int): the number of foreground classes (i.e. background is not
                included). 1 if using class agnostic prediction.
            conv_dims (list[int]): a list of N>0 integers representing the output dimensions
                of N-1 conv layers and the last upsample layer.
            conv_norm (str or callable): normalization for the conv layers.
                See :func:`detectron2.layers.get_norm` for supported types.
        """
        super().__init__(**kwargs)
        assert len(conv_dims) >= 1, "conv_dims have to be non-empty!"

        self.conv_norm_relus = []

        cur_channels = input_shape.channels
        for k, conv_dim in enumerate(conv_dims[:-1]):
            conv = Conv2d(
                cur_channels,
                conv_dim,
                kernel_size=3,
                stride=1,
                padding=1,
                bias=not conv_norm,
                norm=get_norm(conv_norm, conv_dim),
                activation=nn.ReLU(),
            )
            self.add_module("mask_fcn{}".format(k + 1), conv)
            self.conv_norm_relus.append(conv)
            cur_channels = conv_dim

        self.deconv = ConvTranspose2d(
            cur_channels, conv_dims[-1], kernel_size=2, stride=2, padding=0
        )
        self.add_module("deconv_relu", nn.ReLU())
        cur_channels = conv_dims[-1]

        self.predictor = Conv2d(cur_channels, num_classes, kernel_size=1, stride=1, padding=0)

        for layer in self.conv_norm_relus + [self.deconv]:
            weight_init.c2_msra_fill(layer)
        # use normal distribution initialization for mask prediction layer
        nn.init.normal_(self.predictor.weight, std=0.001)
        if self.predictor.bias is not None:
            nn.init.constant_(self.predictor.bias, 0)

    @classmethod
    def from_config(cls, cfg, input_shape):
        ret = super().from_config(cfg, input_shape)
        conv_dim = cfg.MODEL.ROI_MASK_HEAD.CONV_DIM
        num_conv = cfg.MODEL.ROI_MASK_HEAD.NUM_CONV
        ret.update(
            conv_dims=[conv_dim] * (num_conv + 1),  # +1 for ConvTranspose
            conv_norm=cfg.MODEL.ROI_MASK_HEAD.NORM,
            input_shape=input_shape,
        )
        if cfg.MODEL.ROI_MASK_HEAD.CLS_AGNOSTIC_MASK:
            ret["num_classes"] = 1
        else:
            ret["num_classes"] = cfg.MODEL.ROI_HEADS.NUM_CLASSES
        return ret

    def layers(self, x):
        for layer in self:
            x = layer(x)
        return x


def build_mask_head(cfg, input_shape):
    """
    Build a mask head defined by `cfg.MODEL.ROI_MASK_HEAD.NAME`.
    """
    name = cfg.MODEL.ROI_MASK_HEAD.NAME
    return ROI_MASK_HEAD_REGISTRY.get(name)(cfg, input_shape)



# Copyright (c) Facebook, Inc. and its affiliates.
import importlib
import numpy as np
import os
import re
import subprocess
import sys
from collections import defaultdict
import PIL
import torch
import torchvision
from tabulate import tabulate

__all__ = ["collect_env_info"]


def collect_torch_env():
    try:
        import torch.__config__

        return torch.__config__.show()
    except ImportError:
        # compatible with older versions of pytorch
        from torch.utils.collect_env import get_pretty_env_info

        return get_pretty_env_info()


def get_env_module():
    var_name = "DETECTRON2_ENV_MODULE"
    return var_name, os.environ.get(var_name, "<not set>")


def detect_compute_compatibility(CUDA_HOME, so_file):
    try:
        cuobjdump = os.path.join(CUDA_HOME, "bin", "cuobjdump")
        if os.path.isfile(cuobjdump):
            output = subprocess.check_output(
                "'{}' --list-elf '{}'".format(cuobjdump, so_file), shell=True
            )
            output = output.decode("utf-8").strip().split("\n")
            arch = []
            for line in output:
                line = re.findall(r"\.sm_([0-9]*)\.", line)[0]
                arch.append(".".join(line))
            arch = sorted(set(arch))
            return ", ".join(arch)
        else:
            return so_file + "; cannot find cuobjdump"
    except Exception:
        # unhandled failure
        return so_file


def collect_env_info():
    has_gpu = torch.cuda.is_available()  # true for both CUDA & ROCM
    torch_version = torch.__version__

    # NOTE that CUDA_HOME/ROCM_HOME could be None even when CUDA runtime libs are functional
    from torch.utils.cpp_extension import CUDA_HOME, ROCM_HOME

    has_rocm = False
    if (getattr(torch.version, "hip", None) is not None) and (ROCM_HOME is not None):
        has_rocm = True
    has_cuda = has_gpu and (not has_rocm)

    data = []
    data.append(("sys.platform", sys.platform))  # check-template.yml depends on it
    data.append(("Python", sys.version.replace("\n", "")))
    data.append(("numpy", np.__version__))

    try:
        import detectron2  # noqa

        data.append(
            ("detectron2", detectron2.__version__ + " @" + os.path.dirname(detectron2.__file__))
        )
    except ImportError:
        data.append(("detectron2", "failed to import"))
    except AttributeError:
        data.append(("detectron2", "imported a wrong installation"))

    try:
        import detectron2._C as _C
    except ImportError as e:
        data.append(("detectron2._C", f"not built correctly: {e}"))

        # print system compilers when extension fails to build
        if sys.platform != "win32":  # don't know what to do for windows
            try:
                # this is how torch/utils/cpp_extensions.py choose compiler
                cxx = os.environ.get("CXX", "c++")
                cxx = subprocess.check_output("'{}' --version".format(cxx), shell=True)
                cxx = cxx.decode("utf-8").strip().split("\n")[0]
            except subprocess.SubprocessError:
                cxx = "Not found"
            data.append(("Compiler ($CXX)", cxx))

            if has_cuda and CUDA_HOME is not None:
                try:
                    nvcc = os.path.join(CUDA_HOME, "bin", "nvcc")
                    nvcc = subprocess.check_output("'{}' -V".format(nvcc), shell=True)
                    nvcc = nvcc.decode("utf-8").strip().split("\n")[-1]
                except subprocess.SubprocessError:
                    nvcc = "Not found"
                data.append(("CUDA compiler", nvcc))
        if has_cuda and sys.platform != "win32":
            try:
                so_file = importlib.util.find_spec("detectron2._C").origin
            except (ImportError, AttributeError):
                pass
            else:
                data.append(
                    ("detectron2 arch flags", detect_compute_compatibility(CUDA_HOME, so_file))
                )
    else:
        # print compilers that are used to build extension
        data.append(("Compiler", _C.get_compiler_version()))
        data.append(("CUDA compiler", _C.get_cuda_version()))  # cuda or hip
        if has_cuda and getattr(_C, "has_cuda", lambda: True)():
            data.append(
                ("detectron2 arch flags", detect_compute_compatibility(CUDA_HOME, _C.__file__))
            )

    data.append(get_env_module())
    data.append(("PyTorch", torch_version + " @" + os.path.dirname(torch.__file__)))
    data.append(("PyTorch debug build", torch.version.debug))

    if not has_gpu:
        has_gpu_text = "No: torch.cuda.is_available() == False"
    else:
        has_gpu_text = "Yes"
    data.append(("GPU available", has_gpu_text))
    if has_gpu:
        devices = defaultdict(list)
        for k in range(torch.cuda.device_count()):
            cap = ".".join((str(x) for x in torch.cuda.get_device_capability(k)))
            name = torch.cuda.get_device_name(k) + f" (arch={cap})"
            devices[name].append(str(k))
        for name, devids in devices.items():
            data.append(("GPU " + ",".join(devids), name))

        if has_rocm:
            msg = " - invalid!" if not (ROCM_HOME and os.path.isdir(ROCM_HOME)) else ""
            data.append(("ROCM_HOME", str(ROCM_HOME) + msg))
        else:
            try:
                from torch.utils.collect_env import get_nvidia_driver_version, run as _run

                data.append(("Driver version", get_nvidia_driver_version(_run)))
            except Exception:
                pass
            msg = " - invalid!" if not (CUDA_HOME and os.path.isdir(CUDA_HOME)) else ""
            data.append(("CUDA_HOME", str(CUDA_HOME) + msg))

            cuda_arch_list = os.environ.get("TORCH_CUDA_ARCH_LIST", None)
            if cuda_arch_list:
                data.append(("TORCH_CUDA_ARCH_LIST", cuda_arch_list))
    data.append(("Pillow", PIL.__version__))

    try:
        data.append(
            (
                "torchvision",
                str(torchvision.__version__) + " @" + os.path.dirname(torchvision.__file__),
            )
        )
        if has_cuda:
            try:
                torchvision_C = importlib.util.find_spec("torchvision._C").origin
                msg = detect_compute_compatibility(CUDA_HOME, torchvision_C)
                data.append(("torchvision arch flags", msg))
            except (ImportError, AttributeError):
                data.append(("torchvision._C", "Not found"))
    except AttributeError:
        data.append(("torchvision", "unknown"))

    try:
        import fvcore

        data.append(("fvcore", fvcore.__version__))
    except (ImportError, AttributeError):
        pass

    try:
        import iopath

        data.append(("iopath", iopath.__version__))
    except (ImportError, AttributeError):
        pass

    try:
        import cv2

        data.append(("cv2", cv2.__version__))
    except (ImportError, AttributeError):
        data.append(("cv2", "Not found"))
    env_str = tabulate(data) + "\n"
    env_str += collect_torch_env()
    return env_str


def test_nccl_ops():
    num_gpu = torch.cuda.device_count()
    if os.access("/tmp", os.W_OK):
        import torch.multiprocessing as mp

        dist_url = "file:///tmp/nccl_tmp_file"
        print("Testing NCCL connectivity ... this should not hang.")
        mp.spawn(_test_nccl_worker, nprocs=num_gpu, args=(num_gpu, dist_url), daemon=False)
        print("NCCL succeeded.")


def _test_nccl_worker(rank, num_gpu, dist_url):
    import torch.distributed as dist

    dist.init_process_group(backend="NCCL", init_method=dist_url, rank=rank, world_size=num_gpu)
    dist.barrier(device_ids=[rank])


if __name__ == "__main__":
    try:
        from detectron2.utils.collect_env import collect_env_info as f

        print(f())
    except ImportError:
        print(collect_env_info())

    if torch.cuda.is_available():
        num_gpu = torch.cuda.device_count()
        for k in range(num_gpu):
            device = f"cuda:{k}"
            try:
                x = torch.tensor([1, 2.0], dtype=torch.float32)
                x = x.to(device)
            except Exception as e:
                print(
                    f"Unable to copy tensor to device={device}: {e}. "
                    "Your CUDA environment is broken."
                )
        if num_gpu > 1:
            test_nccl_ops()



# Copyright (c) Facebook, Inc. and its affiliates.

import unittest
from typing import List, Sequence, Tuple
import torch

from detectron2.structures import ImageList


class TestImageList(unittest.TestCase):
    def test_imagelist_padding_tracing(self):
        # test that the trace does not contain hard-coded constant sizes
        def to_imagelist(tensors: Sequence[torch.Tensor]):
            image_list = ImageList.from_tensors(tensors, 4)
            return image_list.tensor, image_list.image_sizes

        def _tensor(*shape):
            return torch.ones(shape, dtype=torch.float32)

        # test CHW (inputs needs padding vs. no padding)
        for shape in [(3, 10, 10), (3, 12, 12)]:
            func = torch.jit.trace(to_imagelist, ([_tensor(*shape)],))
            tensor, image_sizes = func([_tensor(3, 15, 20)])
            self.assertEqual(tensor.shape, (1, 3, 16, 20), tensor.shape)
            self.assertEqual(image_sizes[0].tolist(), [15, 20], image_sizes[0])

        # test HW
        func = torch.jit.trace(to_imagelist, ([_tensor(10, 10)],))
        tensor, image_sizes = func([_tensor(15, 20)])
        self.assertEqual(tensor.shape, (1, 16, 20), tensor.shape)
        self.assertEqual(image_sizes[0].tolist(), [15, 20], image_sizes[0])

        # test 2x CHW
        func = torch.jit.trace(
            to_imagelist,
            ([_tensor(3, 16, 10), _tensor(3, 13, 11)],),
        )
        tensor, image_sizes = func([_tensor(3, 25, 20), _tensor(3, 10, 10)])
        self.assertEqual(tensor.shape, (2, 3, 28, 20), tensor.shape)
        self.assertEqual(image_sizes[0].tolist(), [25, 20], image_sizes[0])
        self.assertEqual(image_sizes[1].tolist(), [10, 10], image_sizes[1])
        # support calling with different spatial sizes, but not with different #images

    def test_imagelist_scriptability(self):
        image_nums = 2
        image_tensor = torch.randn((image_nums, 10, 20), dtype=torch.float32)
        image_shape = [(10, 20)] * image_nums

        def f(image_tensor, image_shape: List[Tuple[int, int]]):
            return ImageList(image_tensor, image_shape)

        ret = f(image_tensor, image_shape)
        ret_script = torch.jit.script(f)(image_tensor, image_shape)

        self.assertEqual(len(ret), len(ret_script))
        for i in range(image_nums):
            self.assertTrue(torch.equal(ret[i], ret_script[i]))

    def test_imagelist_from_tensors_scriptability(self):
        image_tensor_0 = torch.randn(10, 20, dtype=torch.float32)
        image_tensor_1 = torch.randn(12, 22, dtype=torch.float32)
        inputs = [image_tensor_0, image_tensor_1]

        def f(image_tensor: List[torch.Tensor]):
            return ImageList.from_tensors(image_tensor, 10)

        ret = f(inputs)
        ret_script = torch.jit.script(f)(inputs)

        self.assertEqual(len(ret), len(ret_script))
        self.assertTrue(torch.equal(ret.tensor, ret_script.tensor))


if __name__ == "__main__":
    unittest.main()



# Copyright 2017 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

r"""Convert raw PASCAL dataset to TFRecord for object_detection.

Example usage:
    python object_detection/dataset_tools/create_pascal_tf_record.py \
        --data_dir=/home/user/VOCdevkit \
        --year=VOC2012 \
        --output_path=/home/user/pascal.record
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import hashlib
import io
import logging
import os

from lxml import etree
import PIL.Image
import tensorflow.compat.v1 as tf

from object_detection.utils import dataset_util
from object_detection.utils import label_map_util


flags = tf.app.flags
flags.DEFINE_string('data_dir', '', 'Root directory to raw PASCAL VOC dataset.')
flags.DEFINE_string('set', 'train', 'Convert training set, validation set or '
                    'merged set.')
flags.DEFINE_string('annotations_dir', 'Annotations',
                    '(Relative) path to annotations directory.')
flags.DEFINE_string('year', 'VOC2007', 'Desired challenge year.')
flags.DEFINE_string('output_path', '', 'Path to output TFRecord')
flags.DEFINE_string('label_map_path', 'data/pascal_label_map.pbtxt',
                    'Path to label map proto')
flags.DEFINE_boolean('ignore_difficult_instances', False, 'Whether to ignore '
                     'difficult instances')
FLAGS = flags.FLAGS

SETS = ['train', 'val', 'trainval', 'test']
YEARS = ['VOC2007', 'VOC2012', 'merged']


def dict_to_tf_example(data,
                       dataset_directory,
                       label_map_dict,
                       ignore_difficult_instances=False,
                       image_subdirectory='JPEGImages'):
  """Convert XML derived dict to tf.Example proto.

  Notice that this function normalizes the bounding box coordinates provided
  by the raw data.

  Args:
    data: dict holding PASCAL XML fields for a single image (obtained by
      running dataset_util.recursive_parse_xml_to_dict)
    dataset_directory: Path to root directory holding PASCAL dataset
    label_map_dict: A map from string label names to integers ids.
    ignore_difficult_instances: Whether to skip difficult instances in the
      dataset  (default: False).
    image_subdirectory: String specifying subdirectory within the
      PASCAL dataset directory holding the actual image data.

  Returns:
    example: The converted tf.Example.

  Raises:
    ValueError: if the image pointed to by data['filename'] is not a valid JPEG
  """
  img_path = os.path.join(data['folder'], image_subdirectory, data['filename'])
  full_path = os.path.join(dataset_directory, img_path)
  with tf.gfile.GFile(full_path, 'rb') as fid:
    encoded_jpg = fid.read()
  encoded_jpg_io = io.BytesIO(encoded_jpg)
  image = PIL.Image.open(encoded_jpg_io)
  if image.format != 'JPEG':
    raise ValueError('Image format not JPEG')
  key = hashlib.sha256(encoded_jpg).hexdigest()

  width = int(data['size']['width'])
  height = int(data['size']['height'])

  xmin = []
  ymin = []
  xmax = []
  ymax = []
  classes = []
  classes_text = []
  truncated = []
  poses = []
  difficult_obj = []
  if 'object' in data:
    for obj in data['object']:
      difficult = bool(int(obj['difficult']))
      if ignore_difficult_instances and difficult:
        continue

      difficult_obj.append(int(difficult))

      xmin.append(float(obj['bndbox']['xmin']) / width)
      ymin.append(float(obj['bndbox']['ymin']) / height)
      xmax.append(float(obj['bndbox']['xmax']) / width)
      ymax.append(float(obj['bndbox']['ymax']) / height)
      classes_text.append(obj['name'].encode('utf8'))
      classes.append(label_map_dict[obj['name']])
      truncated.append(int(obj['truncated']))
      poses.append(obj['pose'].encode('utf8'))

  example = tf.train.Example(features=tf.train.Features(feature={
      'image/height': dataset_util.int64_feature(height),
      'image/width': dataset_util.int64_feature(width),
      'image/filename': dataset_util.bytes_feature(
          data['filename'].encode('utf8')),
      'image/source_id': dataset_util.bytes_feature(
          data['filename'].encode('utf8')),
      'image/key/sha256': dataset_util.bytes_feature(key.encode('utf8')),
      'image/encoded': dataset_util.bytes_feature(encoded_jpg),
      'image/format': dataset_util.bytes_feature('jpeg'.encode('utf8')),
      'image/object/bbox/xmin': dataset_util.float_list_feature(xmin),
      'image/object/bbox/xmax': dataset_util.float_list_feature(xmax),
      'image/object/bbox/ymin': dataset_util.float_list_feature(ymin),
      'image/object/bbox/ymax': dataset_util.float_list_feature(ymax),
      'image/object/class/text': dataset_util.bytes_list_feature(classes_text),
      'image/object/class/label': dataset_util.int64_list_feature(classes),
      'image/object/difficult': dataset_util.int64_list_feature(difficult_obj),
      'image/object/truncated': dataset_util.int64_list_feature(truncated),
      'image/object/view': dataset_util.bytes_list_feature(poses),
  }))
  return example


def main(_):
  if FLAGS.set not in SETS:
    raise ValueError('set must be in : {}'.format(SETS))
  if FLAGS.year not in YEARS:
    raise ValueError('year must be in : {}'.format(YEARS))

  data_dir = FLAGS.data_dir
  years = ['VOC2007', 'VOC2012']
  if FLAGS.year != 'merged':
    years = [FLAGS.year]

  writer = tf.python_io.TFRecordWriter(FLAGS.output_path)

  label_map_dict = label_map_util.get_label_map_dict(FLAGS.label_map_path)

  for year in years:
    logging.info('Reading from PASCAL %s dataset.', year)
    examples_path = os.path.join(data_dir, year, 'ImageSets', 'Main',
                                 'aeroplane_' + FLAGS.set + '.txt')
    annotations_dir = os.path.join(data_dir, year, FLAGS.annotations_dir)
    examples_list = dataset_util.read_examples_list(examples_path)
    for idx, example in enumerate(examples_list):
      if idx % 100 == 0:
        logging.info('On image %d of %d', idx, len(examples_list))
      path = os.path.join(annotations_dir, example + '.xml')
      with tf.gfile.GFile(path, 'r') as fid:
        xml_str = fid.read()
      xml = etree.fromstring(xml_str)
      data = dataset_util.recursive_parse_xml_to_dict(xml)['annotation']

      tf_example = dict_to_tf_example(data, FLAGS.data_dir, label_map_dict,
                                      FLAGS.ignore_difficult_instances)
      writer.write(tf_example.SerializeToString())

  writer.close()


if __name__ == '__main__':
  tf.app.run()



#!/usr/bin/env python
# Copyright (c) Facebook, Inc. and its affiliates.
"""
Detectron2 training script with a plain training loop.

This script reads a given config file and runs the training or evaluation.
It is an entry point that is able to train standard models in detectron2.

In order to let one script support training of many models,
this script contains logic that are specific to these built-in models and therefore
may not be suitable for your own project.
For example, your research project perhaps only needs a single "evaluator".

Therefore, we recommend you to use detectron2 as a library and take
this file as an example of how to use the library.
You may want to write your own script with your datasets and other customizations.

Compared to "train_net.py", this script supports fewer default features.
It also includes fewer abstraction, therefore is easier to add custom logic.
"""

import logging
import os
from collections import OrderedDict
import torch
from torch.nn.parallel import DistributedDataParallel

import detectron2.utils.comm as comm
from detectron2.checkpoint import DetectionCheckpointer, PeriodicCheckpointer
from detectron2.config import get_cfg
from detectron2.data import (
    MetadataCatalog,
    build_detection_test_loader,
    build_detection_train_loader,
)
from detectron2.engine import default_argument_parser, default_setup, default_writers, launch
from detectron2.evaluation import (
    CityscapesInstanceEvaluator,
    CityscapesSemSegEvaluator,
    COCOEvaluator,
    COCOPanopticEvaluator,
    DatasetEvaluators,
    LVISEvaluator,
    PascalVOCDetectionEvaluator,
    SemSegEvaluator,
    inference_on_dataset,
    print_csv_format,
)
from detectron2.modeling import build_model
from detectron2.solver import build_lr_scheduler, build_optimizer
from detectron2.utils.events import EventStorage

logger = logging.getLogger("detectron2")


def get_evaluator(cfg, dataset_name, output_folder=None):
    """
    Create evaluator(s) for a given dataset.
    This uses the special metadata "evaluator_type" associated with each builtin dataset.
    For your own dataset, you can simply create an evaluator manually in your
    script and do not have to worry about the hacky if-else logic here.
    """
    if output_folder is None:
        output_folder = os.path.join(cfg.OUTPUT_DIR, "inference")
    evaluator_list = []
    evaluator_type = MetadataCatalog.get(dataset_name).evaluator_type
    if evaluator_type in ["sem_seg", "coco_panoptic_seg"]:
        evaluator_list.append(
            SemSegEvaluator(
                dataset_name,
                distributed=True,
                output_dir=output_folder,
            )
        )
    if evaluator_type in ["coco", "coco_panoptic_seg"]:
        evaluator_list.append(COCOEvaluator(dataset_name, output_dir=output_folder))
    if evaluator_type == "coco_panoptic_seg":
        evaluator_list.append(COCOPanopticEvaluator(dataset_name, output_folder))
    if evaluator_type == "cityscapes_instance":
        return CityscapesInstanceEvaluator(dataset_name)
    if evaluator_type == "cityscapes_sem_seg":
        return CityscapesSemSegEvaluator(dataset_name)
    if evaluator_type == "pascal_voc":
        return PascalVOCDetectionEvaluator(dataset_name)
    if evaluator_type == "lvis":
        return LVISEvaluator(dataset_name, cfg, True, output_folder)
    if len(evaluator_list) == 0:
        raise NotImplementedError(
            "no Evaluator for the dataset {} with the type {}".format(dataset_name, evaluator_type)
        )
    if len(evaluator_list) == 1:
        return evaluator_list[0]
    return DatasetEvaluators(evaluator_list)


def do_test(cfg, model):
    results = OrderedDict()
    for dataset_name in cfg.DATASETS.TEST:
        data_loader = build_detection_test_loader(cfg, dataset_name)
        evaluator = get_evaluator(
            cfg, dataset_name, os.path.join(cfg.OUTPUT_DIR, "inference", dataset_name)
        )
        results_i = inference_on_dataset(model, data_loader, evaluator)
        results[dataset_name] = results_i
        if comm.is_main_process():
            logger.info("Evaluation results for {} in csv format:".format(dataset_name))
            print_csv_format(results_i)
    if len(results) == 1:
        results = list(results.values())[0]
    return results


def do_train(cfg, model, resume=False):
    model.train()
    optimizer = build_optimizer(cfg, model)
    scheduler = build_lr_scheduler(cfg, optimizer)

    checkpointer = DetectionCheckpointer(
        model, cfg.OUTPUT_DIR, optimizer=optimizer, scheduler=scheduler
    )
    start_iter = (
        checkpointer.resume_or_load(cfg.MODEL.WEIGHTS, resume=resume).get("iteration", -1) + 1
    )
    max_iter = cfg.SOLVER.MAX_ITER

    periodic_checkpointer = PeriodicCheckpointer(
        checkpointer, cfg.SOLVER.CHECKPOINT_PERIOD, max_iter=max_iter
    )

    writers = default_writers(cfg.OUTPUT_DIR, max_iter) if comm.is_main_process() else []

    # compared to "train_net.py", we do not support accurate timing and
    # precise BN here, because they are not trivial to implement in a small training loop
    data_loader = build_detection_train_loader(cfg)
    logger.info("Starting training from iteration {}".format(start_iter))
    with EventStorage(start_iter) as storage:
        for data, iteration in zip(data_loader, range(start_iter, max_iter)):
            storage.iter = iteration

            loss_dict = model(data)
            losses = sum(loss_dict.values())
            assert torch.isfinite(losses).all(), loss_dict

            loss_dict_reduced = {k: v.item() for k, v in comm.reduce_dict(loss_dict).items()}
            losses_reduced = sum(loss for loss in loss_dict_reduced.values())
            if comm.is_main_process():
                storage.put_scalars(total_loss=losses_reduced, **loss_dict_reduced)

            optimizer.zero_grad()
            losses.backward()
            optimizer.step()
            storage.put_scalar("lr", optimizer.param_groups[0]["lr"], smoothing_hint=False)
            scheduler.step()

            if (
                cfg.TEST.EVAL_PERIOD > 0
                and (iteration + 1) % cfg.TEST.EVAL_PERIOD == 0
                and iteration != max_iter - 1
            ):
                do_test(cfg, model)
                # Compared to "train_net.py", the test results are not dumped to EventStorage
                comm.synchronize()

            if iteration - start_iter > 5 and (
                (iteration + 1) % 20 == 0 or iteration == max_iter - 1
            ):
                for writer in writers:
                    writer.write()
            periodic_checkpointer.step(iteration)


def setup(args):
    """
    Create configs and perform basic setups.
    """
    cfg = get_cfg()
    cfg.merge_from_file(args.config_file)
    cfg.merge_from_list(args.opts)
    cfg.freeze()
    default_setup(
        cfg, args
    )  # if you don't like any of the default setup, write your own setup code
    return cfg


def main(args):
    cfg = setup(args)

    model = build_model(cfg)
    logger.info("Model:\n{}".format(model))
    if args.eval_only:
        DetectionCheckpointer(model, save_dir=cfg.OUTPUT_DIR).resume_or_load(
            cfg.MODEL.WEIGHTS, resume=args.resume
        )
        return do_test(cfg, model)

    distributed = comm.get_world_size() > 1
    if distributed:
        model = DistributedDataParallel(
            model, device_ids=[comm.get_local_rank()], broadcast_buffers=False
        )

    do_train(cfg, model, resume=args.resume)
    return do_test(cfg, model)


if __name__ == "__main__":
    args = default_argument_parser().parse_args()
    print("Command Line Args:", args)
    launch(
        main,
        args.num_gpus,
        num_machines=args.num_machines,
        machine_rank=args.machine_rank,
        dist_url=args.dist_url,
        args=(args,),
    )



# Copyright 2017 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Tests for tensorflow_models.object_detection.utils.vrd_evaluation."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import numpy as np
import tensorflow.compat.v1 as tf

from object_detection.core import standard_fields
from object_detection.utils import vrd_evaluation


class VRDRelationDetectionEvaluatorTest(tf.test.TestCase):

  def test_vrdrelation_evaluator(self):
    self.vrd_eval = vrd_evaluation.VRDRelationDetectionEvaluator()

    image_key1 = 'img1'
    groundtruth_box_tuples1 = np.array(
        [([0, 0, 1, 1], [1, 1, 2, 2]), ([0, 0, 1, 1], [1, 2, 2, 3])],
        dtype=vrd_evaluation.vrd_box_data_type)
    groundtruth_class_tuples1 = np.array(
        [(1, 2, 3), (1, 4, 3)], dtype=vrd_evaluation.label_data_type)
    groundtruth_verified_labels1 = np.array([1, 2, 3, 4, 5], dtype=int)
    self.vrd_eval.add_single_ground_truth_image_info(
        image_key1, {
            standard_fields.InputDataFields.groundtruth_boxes:
                groundtruth_box_tuples1,
            standard_fields.InputDataFields.groundtruth_classes:
                groundtruth_class_tuples1,
            standard_fields.InputDataFields.groundtruth_image_classes:
                groundtruth_verified_labels1
        })

    image_key2 = 'img2'
    groundtruth_box_tuples2 = np.array(
        [([0, 0, 1, 1], [1, 1, 2, 2])], dtype=vrd_evaluation.vrd_box_data_type)
    groundtruth_class_tuples2 = np.array(
        [(1, 4, 3)], dtype=vrd_evaluation.label_data_type)
    self.vrd_eval.add_single_ground_truth_image_info(
        image_key2, {
            standard_fields.InputDataFields.groundtruth_boxes:
                groundtruth_box_tuples2,
            standard_fields.InputDataFields.groundtruth_classes:
                groundtruth_class_tuples2,
        })

    image_key3 = 'img3'
    groundtruth_box_tuples3 = np.array(
        [([0, 0, 1, 1], [1, 1, 2, 2])], dtype=vrd_evaluation.vrd_box_data_type)
    groundtruth_class_tuples3 = np.array(
        [(1, 2, 4)], dtype=vrd_evaluation.label_data_type)
    self.vrd_eval.add_single_ground_truth_image_info(
        image_key3, {
            standard_fields.InputDataFields.groundtruth_boxes:
                groundtruth_box_tuples3,
            standard_fields.InputDataFields.groundtruth_classes:
                groundtruth_class_tuples3,
        })

    image_key = 'img1'
    detected_box_tuples = np.array(
        [([0, 0.3, 1, 1], [1.1, 1, 2, 2]), ([0, 0, 1, 1], [1, 1, 2, 2]),
         ([0.5, 0, 1, 1], [1, 1, 3, 3])],
        dtype=vrd_evaluation.vrd_box_data_type)
    detected_class_tuples = np.array(
        [(1, 2, 5), (1, 2, 3), (1, 6, 3)], dtype=vrd_evaluation.label_data_type)
    detected_scores = np.array([0.7, 0.8, 0.9], dtype=float)
    self.vrd_eval.add_single_detected_image_info(
        image_key, {
            standard_fields.DetectionResultFields.detection_boxes:
                detected_box_tuples,
            standard_fields.DetectionResultFields.detection_scores:
                detected_scores,
            standard_fields.DetectionResultFields.detection_classes:
                detected_class_tuples
        })

    metrics = self.vrd_eval.evaluate()

    self.assertAlmostEqual(metrics['VRDMetric_Relationships_weightedAP@0.5IOU'],
                           0.25)
    self.assertAlmostEqual(metrics['VRDMetric_Relationships_mAP@0.5IOU'],
                           0.1666666666666666)
    self.assertAlmostEqual(metrics['VRDMetric_Relationships_AP@0.5IOU/3'],
                           0.3333333333333333)
    self.assertAlmostEqual(metrics['VRDMetric_Relationships_AP@0.5IOU/4'], 0)
    self.assertAlmostEqual(metrics['VRDMetric_Relationships_Recall@50@0.5IOU'],
                           0.25)
    self.assertAlmostEqual(metrics['VRDMetric_Relationships_Recall@100@0.5IOU'],
                           0.25)
    self.vrd_eval.clear()
    self.assertFalse(self.vrd_eval._image_ids)


class VRDPhraseDetectionEvaluatorTest(tf.test.TestCase):

  def test_vrdphrase_evaluator(self):
    self.vrd_eval = vrd_evaluation.VRDPhraseDetectionEvaluator()

    image_key1 = 'img1'
    groundtruth_box_tuples1 = np.array(
        [([0, 0, 1, 1], [1, 1, 2, 2]), ([0, 0, 1, 1], [1, 2, 2, 3])],
        dtype=vrd_evaluation.vrd_box_data_type)
    groundtruth_class_tuples1 = np.array(
        [(1, 2, 3), (1, 4, 3)], dtype=vrd_evaluation.label_data_type)
    groundtruth_verified_labels1 = np.array([1, 2, 3, 4, 5], dtype=int)
    self.vrd_eval.add_single_ground_truth_image_info(
        image_key1, {
            standard_fields.InputDataFields.groundtruth_boxes:
                groundtruth_box_tuples1,
            standard_fields.InputDataFields.groundtruth_classes:
                groundtruth_class_tuples1,
            standard_fields.InputDataFields.groundtruth_image_classes:
                groundtruth_verified_labels1
        })

    image_key2 = 'img2'
    groundtruth_box_tuples2 = np.array(
        [([0, 0, 1, 1], [1, 1, 2, 2])], dtype=vrd_evaluation.vrd_box_data_type)
    groundtruth_class_tuples2 = np.array(
        [(1, 4, 3)], dtype=vrd_evaluation.label_data_type)
    self.vrd_eval.add_single_ground_truth_image_info(
        image_key2, {
            standard_fields.InputDataFields.groundtruth_boxes:
                groundtruth_box_tuples2,
            standard_fields.InputDataFields.groundtruth_classes:
                groundtruth_class_tuples2,
        })

    image_key3 = 'img3'
    groundtruth_box_tuples3 = np.array(
        [([0, 0, 1, 1], [1, 1, 2, 2])], dtype=vrd_evaluation.vrd_box_data_type)
    groundtruth_class_tuples3 = np.array(
        [(1, 2, 4)], dtype=vrd_evaluation.label_data_type)
    self.vrd_eval.add_single_ground_truth_image_info(
        image_key3, {
            standard_fields.InputDataFields.groundtruth_boxes:
                groundtruth_box_tuples3,
            standard_fields.InputDataFields.groundtruth_classes:
                groundtruth_class_tuples3,
        })

    image_key = 'img1'
    detected_box_tuples = np.array(
        [([0, 0.3, 0.5, 0.5], [0.3, 0.3, 1.0, 1.0]),
         ([0, 0, 1.2, 1.2], [0.0, 0.0, 2.0, 2.0]),
         ([0.5, 0, 1, 1], [1, 1, 3, 3])],
        dtype=vrd_evaluation.vrd_box_data_type)
    detected_class_tuples = np.array(
        [(1, 2, 5), (1, 2, 3), (1, 6, 3)], dtype=vrd_evaluation.label_data_type)
    detected_scores = np.array([0.7, 0.8, 0.9], dtype=float)
    self.vrd_eval.add_single_detected_image_info(
        image_key, {
            standard_fields.DetectionResultFields.detection_boxes:
                detected_box_tuples,
            standard_fields.DetectionResultFields.detection_scores:
                detected_scores,
            standard_fields.DetectionResultFields.detection_classes:
                detected_class_tuples
        })

    metrics = self.vrd_eval.evaluate()

    self.assertAlmostEqual(metrics['VRDMetric_Phrases_weightedAP@0.5IOU'], 0.25)
    self.assertAlmostEqual(metrics['VRDMetric_Phrases_mAP@0.5IOU'],
                           0.1666666666666666)
    self.assertAlmostEqual(metrics['VRDMetric_Phrases_AP@0.5IOU/3'],
                           0.3333333333333333)
    self.assertAlmostEqual(metrics['VRDMetric_Phrases_AP@0.5IOU/4'], 0)
    self.assertAlmostEqual(metrics['VRDMetric_Phrases_Recall@50@0.5IOU'], 0.25)
    self.assertAlmostEqual(metrics['VRDMetric_Phrases_Recall@100@0.5IOU'], 0.25)

    self.vrd_eval.clear()
    self.assertFalse(self.vrd_eval._image_ids)


class VRDDetectionEvaluationTest(tf.test.TestCase):

  def setUp(self):

    self.vrd_eval = vrd_evaluation._VRDDetectionEvaluation(
        matching_iou_threshold=0.5)

    image_key1 = 'img1'
    groundtruth_box_tuples1 = np.array(
        [([0, 0, 1, 1], [1, 1, 2, 2]), ([0, 0, 1, 1], [1, 2, 2, 3])],
        dtype=vrd_evaluation.vrd_box_data_type)
    groundtruth_class_tuples1 = np.array(
        [(1, 2, 3), (1, 4, 3)], dtype=vrd_evaluation.label_data_type)
    self.vrd_eval.add_single_ground_truth_image_info(
        image_key1, groundtruth_box_tuples1, groundtruth_class_tuples1)

    image_key2 = 'img2'
    groundtruth_box_tuples2 = np.array(
        [([0, 0, 1, 1], [1, 1, 2, 2])], dtype=vrd_evaluation.vrd_box_data_type)
    groundtruth_class_tuples2 = np.array(
        [(1, 4, 3)], dtype=vrd_evaluation.label_data_type)
    self.vrd_eval.add_single_ground_truth_image_info(
        image_key2, groundtruth_box_tuples2, groundtruth_class_tuples2)

    image_key3 = 'img3'
    groundtruth_box_tuples3 = np.array(
        [([0, 0, 1, 1], [1, 1, 2, 2])], dtype=vrd_evaluation.vrd_box_data_type)
    groundtruth_class_tuples3 = np.array(
        [(1, 2, 4)], dtype=vrd_evaluation.label_data_type)
    self.vrd_eval.add_single_ground_truth_image_info(
        image_key3, groundtruth_box_tuples3, groundtruth_class_tuples3)

    image_key = 'img1'
    detected_box_tuples = np.array(
        [([0, 0.3, 1, 1], [1.1, 1, 2, 2]), ([0, 0, 1, 1], [1, 1, 2, 2])],
        dtype=vrd_evaluation.vrd_box_data_type)
    detected_class_tuples = np.array(
        [(1, 2, 3), (1, 2, 3)], dtype=vrd_evaluation.label_data_type)
    detected_scores = np.array([0.7, 0.8], dtype=float)
    self.vrd_eval.add_single_detected_image_info(
        image_key, detected_box_tuples, detected_scores, detected_class_tuples)

    metrics = self.vrd_eval.evaluate()
    expected_weighted_average_precision = 0.25
    expected_mean_average_precision = 0.16666666666666
    expected_precision = np.array([1., 0.5], dtype=float)
    expected_recall = np.array([0.25, 0.25], dtype=float)
    expected_recall_50 = 0.25
    expected_recall_100 = 0.25
    expected_median_rank_50 = 0
    expected_median_rank_100 = 0

    self.assertAlmostEqual(expected_weighted_average_precision,
                           metrics.weighted_average_precision)
    self.assertAlmostEqual(expected_mean_average_precision,
                           metrics.mean_average_precision)
    self.assertAlmostEqual(expected_mean_average_precision,
                           metrics.mean_average_precision)

    self.assertAllClose(expected_precision, metrics.precisions)
    self.assertAllClose(expected_recall, metrics.recalls)
    self.assertAlmostEqual(expected_recall_50, metrics.recall_50)
    self.assertAlmostEqual(expected_recall_100, metrics.recall_100)
    self.assertAlmostEqual(expected_median_rank_50, metrics.median_rank_50)
    self.assertAlmostEqual(expected_median_rank_100, metrics.median_rank_100)


if __name__ == '__main__':
  tf.test.main()
